#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import copy
import datetime
import json
import logging
import sys

import pandas as pd
from collections import namedtuple
from decimal import Decimal

from sqlalchemy import func, null, and_, or_, distinct
from sqlalchemy.orm import aliased, make_transient, joinedload
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import literal_column, case

from .objects import JSONObject, Dataset as JSONDataset
from .. import db
from .. import config
from ..db.model import Dataset, Metadata, DatasetOwner, DatasetCollection,\
        DatasetCollectionItem, ResourceScenario, ResourceAttr, TypeAttr
from ..exceptions import HydraError, PermissionError, ResourceNotFoundError
from ..util import generate_data_hash
from ..util.hydra_dateutil import get_datetime
from ..util.permissions import required_role

from hydra_base.lib.storage import (
    MongoStorageAdapter,
    HdfStorageAdapter
)


global FORMAT
FORMAT = "%Y-%m-%d %H:%M:%S.%f"
global qry_in_threshold
qry_in_threshold = 999
#"2013-08-13T15:55:43.468886Z"

current_module = sys.modules[__name__]
NS = "soap_server.hydra_complexmodels"

log = logging.getLogger(__name__)

def get_dataset(dataset_id,**kwargs):
    """
        Get a single dataset, by ID
    """

    user_id = int(kwargs.get('user_id'))

    if dataset_id is None:
        return None
    try:
        dataset_rs = db.DBSession.query(Dataset.id,
                Dataset.type,
                Dataset.unit_id,
                Dataset.name,
                Dataset.hidden,
                Dataset.cr_date,
                Dataset.created_by,
                DatasetOwner.user_id,
                null().label('metadata'),
                case((and_(Dataset.hidden=='Y', DatasetOwner.user_id is not None), None),
                        else_=Dataset.value).label('value')).filter(
                Dataset.id==dataset_id).outerjoin(DatasetOwner,
                                    and_(DatasetOwner.dataset_id==Dataset.id,
                                    DatasetOwner.user_id==user_id)).one()

        rs_dict = dataset_rs._asdict()

        #convert the value row into a string as it is returned as a binary

        if dataset_rs.value is not None:
            rs_dict['value'] = str(dataset_rs.value)

        if dataset_rs.hidden == 'N' or (dataset_rs.hidden == 'Y' and dataset_rs.user_id is not None):
            metadata = db.DBSession.query(Metadata).filter(Metadata.dataset_id==dataset_id).all()
            rs_dict['metadata'] = metadata
        else:
            rs_dict['metadata'] = []

    except NoResultFound:
        raise HydraError("Dataset %s does not exist."%(dataset_id))

    dataset = JSONDataset(rs_dict)

    return dataset

def clone_dataset(dataset_id,**kwargs):
    """
        Get a single dataset, by ID
    """

    user_id = int(kwargs.get('user_id'))

    if dataset_id is None:
        return None

    dataset = db.DBSession.query(Dataset).filter(
            Dataset.id==dataset_id).options(joinedload(Dataset.metadata)).first()

    if dataset is None:
        raise HydraError("Dataset %s does not exist."%(dataset_id))

    if dataset is not None and dataset.created_by != user_id:
        owner = db.DBSession.query(DatasetOwner).filter(
                                DatasetOwner.dataset_id==Dataset.id,
                                DatasetOwner.user_id==user_id).first()
        if owner is None:
            raise PermissionError("User %s is not an owner of dataset %s and therefore cannot clone it."%(user_id, dataset_id))

    db.DBSession.expunge(dataset)

    make_transient(dataset)

    dataset.name = dataset.name + "(Clone)"
    dataset.id = None
    dataset.cr_date = None

    #Try to avoid duplicate metadata entries if the entry has been cloned previously
    for m in dataset.metadata:
        if m.key in ("clone_of", "cloned_by"):
            del(m)

    cloned_meta = Metadata()
    cloned_meta.key = "clone_of"
    cloned_meta.value  = str(dataset_id)
    dataset.metadata.append(cloned_meta)
    cloned_meta = Metadata()
    cloned_meta.key = "cloned_by"
    cloned_meta.value  = str(user_id)
    dataset.metadata.append(cloned_meta)

    dataset.set_hash()
    db.DBSession.add(dataset)
    db.DBSession.flush()

    cloned_dataset = db.DBSession.query(Dataset).filter(
            Dataset.id==dataset.id).first()

    return cloned_dataset

def get_datasets(dataset_ids,**kwargs):
    """
        Get a single dataset, by ID
    """

    user_id = int(kwargs.get('user_id'))
    datasets = []
    if len(dataset_ids) == 0:
        return []
    try:
        dataset_rs = db.DBSession.query(Dataset.id,
                Dataset.type,
                Dataset.unit_id,
                Dataset.name,
                Dataset.hidden,
                Dataset.cr_date,
                Dataset.created_by,
                DatasetOwner.user_id,
                null().label('metadata'),
                case([(and_(Dataset.hidden=='Y', DatasetOwner.user_id is not None), None)],
                        else_=Dataset.value).label('value')).filter(
                Dataset.id.in_(dataset_ids)).outerjoin(DatasetOwner,
                                    and_(DatasetOwner.dataset_id==Dataset.id,
                                    DatasetOwner.user_id==user_id)).all()

        #convert the value row into a string as it is returned as a binary
        for dataset_row in dataset_rs:
            dataset_dict = dataset_row._asdict()

            if dataset_row.value is not None:
                dataset_dict['value'] = str(dataset_row.value)

            if dataset_row.hidden == 'N' or (dataset_row.hidden == 'Y' and dataset_row.user_id is not None):
                metadata = db.DBSession.query(Metadata).filter(Metadata.dataset_id == dataset_row.id).all()
                dataset_dict['metadata'] = metadata
            else:
                dataset_dict['metadata'] = []

            datasets.append(namedtuple('Dataset', dataset_dict.keys())(**dataset_dict))


    except NoResultFound:
        raise ResourceNotFoundError("Datasets not found.")

    return datasets



def search_datasets(dataset_id=None,
                dataset_name=None,
                collection_name=None,
                data_type=None,
                unit_id=None,
                scenario_id=None,
                metadata_key=None,
                metadata_val=None,
                attr_id = None,
                type_id = None,
                unconnected = None,
                inc_metadata='N',
                inc_val = 'N',
                page_start = 0,
                page_size   = 2000,
                **kwargs):
    """
        Get multiple datasets, based on several
        filters. If all filters are set to None, all
        datasets in the DB (that the user is allowe to see)
        will be returned.
    """


    log.info("Searching datasets: \ndatset_id: %s,\n"
                                  "datset_name: %s,\n"
                                  "collection_name: %s,\n"
                                  "data_type: %s,\n"
                                  "unit_id: %s,\n"
                                  "scenario_id: %s,\n"
                                  "metadata_key: %s,\n"
                                  "metadata_val: %s,\n"
                                  "attr_id: %s,\n"
                                  "type_id: %s,\n"
                                  "unconnected: %s,\n"
                                  "inc_metadata: %s,\n"
                                  "inc_val: %s,\n"
                                  "page_start: %s,\n"
                                  "page_size: %s" % (dataset_id,
                dataset_name,
                collection_name,
                data_type,
                unit_id,
                scenario_id,
                metadata_key,
                metadata_val,
                attr_id,
                type_id,
                unconnected,
                inc_metadata,
                inc_val,
                page_start,
                page_size))

    if page_size is None:
        page_size = config.get('SEARCH', 'page_size', 2000)

    user_id = int(kwargs.get('user_id'))

    dataset_qry = db.DBSession.query(Dataset.id,
            Dataset.type,
            Dataset.unit_id,
            Dataset.name,
            Dataset.hidden,
            Dataset.cr_date,
            Dataset.created_by,
            DatasetOwner.user_id,
            null().label('metadata'),
            Dataset.value
    )

    #Dataset ID is unique, so there's no point using the other filters.
    #Only use other filters if the datset ID is not specified.
    if dataset_id is not None:
        dataset_qry = dataset_qry.filter(
            Dataset.id==dataset_id)

    else:
        if dataset_name is not None:
            dataset_qry = dataset_qry.filter(
                func.lower(Dataset.name).like("%%%s%%"%dataset_name.lower())
            )
        if collection_name is not None:
            dc = aliased(DatasetCollection)
            dci = aliased(DatasetCollectionItem)
            dataset_qry = dataset_qry.join(dc,
                        func.lower(dc.name).like("%%%s%%"%collection_name.lower())
                        ).join(dci,and_(
                            dci.collection_id == dc.id,
                            dci.dataset_id == Dataset.id))

        if data_type is not None:
            dataset_qry = dataset_qry.filter(
                func.lower(Dataset.type) == data_type.lower())

        #null is a valid unit, so we need a way for the searcher
        #to specify that they want to search for datasets with a null unit
        #rather than ignoring the unit. We use 'null' to do this.

        if unit_id is not None:
            dataset_qry = dataset_qry.filter(
                Dataset.unit_id == unit_id)

        if scenario_id is not None:
            dataset_qry = dataset_qry.join(ResourceScenario,
                                and_(ResourceScenario.dataset_id == Dataset.id,
                                ResourceScenario.scenario_id == scenario_id))

        if attr_id is not None:
            dataset_qry = dataset_qry.join(
                ResourceScenario, ResourceScenario.dataset_id == Dataset.id).join(
                ResourceAttr, and_(ResourceAttr.id==ResourceScenario.resource_attr_id,
                                  ResourceAttr.attr_id==attr_id))

        if type_id is not None:
            dataset_qry = dataset_qry.join(
                ResourceScenario, ResourceScenario.dataset_id == Dataset.id).join(
                ResourceAttr, ResourceAttr.id==ResourceScenario.resource_attr_id).join(
                TypeAttr, and_(TypeAttr.attr_id==ResourceAttr.attr_id, TypeAttr.type_id==type_id))

        if unconnected == 'Y':
            stmt = db.DBSession.query(distinct(ResourceScenario.dataset_id).label('dataset_id'),
                                literal_column("0").label('col')).subquery()
            dataset_qry = dataset_qry.outerjoin(
                stmt, stmt.c.dataset_id == Dataset.id)
            dataset_qry = dataset_qry.filter(stmt.c.col == None)
        elif unconnected == 'N':
            #The dataset has to be connected to something
            stmt = db.DBSession.query(distinct(ResourceScenario.dataset_id).label('dataset_id'),
                                literal_column("0").label('col')).subquery()
            dataset_qry = dataset_qry.join(
                stmt, stmt.c.dataset_id == Dataset.id)
        if metadata_key is not None and metadata_val is not None:
            dataset_qry = dataset_qry.join(Metadata,
                                and_(Metadata.dataset_id == Dataset.id,
                                func.lower(Metadata.key).like("%%%s%%"%metadata_key.lower()),
                                func.lower(Metadata.value).like("%%%s%%"%metadata_val.lower())))
        elif metadata_key is not None and metadata_val is None:
            dataset_qry = dataset_qry.join(Metadata,
                                and_(Metadata.dataset_id == Dataset.id,
                                func.lower(Metadata.key).like("%%%s%%"%metadata_key.lower())))
        elif metadata_key is None and metadata_val is not None:
            dataset_qry = dataset_qry.join(Metadata,
                                and_(Metadata.dataset_id == Dataset.id,
                                func.lower(Metadata.value).like("%%%s%%"%metadata_val.lower())))

    #All datasets must be joined on dataset owner so only datasets that the
    #user can see are retrieved.
    dataset_qry = dataset_qry.outerjoin(DatasetOwner,
                                and_(DatasetOwner.dataset_id==Dataset.id,
                                DatasetOwner.user_id==user_id))

    dataset_qry = dataset_qry.filter(or_(Dataset.hidden=='N', and_(DatasetOwner.user_id is not None, Dataset.hidden=='Y')))

    log.info(str(dataset_qry))

    datasets = dataset_qry.all()

    log.info("Retrieved %s datasets", len(datasets))

    #page the datasets:
    if page_start + page_size > len(datasets):
        page_end = None
    else:
        page_end = page_start + page_size

    datasets = datasets[page_start:page_end]

    log.info("Datasets paged from result %s to %s", page_start, page_end)

    datasets_to_return = []
    for dataset_row in datasets:

        dataset_dict = dataset_row._asdict()



        if inc_val == 'N':
            dataset_dict['value'] = None
        else:
            #convert the value row into a string as it is returned as a binary
            if dataset_row.value is not None:
                dataset_dict['value'] = str(dataset_row.value)

        if inc_metadata=='Y':
            metadata = db.DBSession.query(Metadata).filter(Metadata.dataset_id==dataset_row.dataset_id).all()
            dataset_dict['metadata'] = metadata
        else:
            dataset_dict['metadata'] = []

        dataset = namedtuple('Dataset', dataset_dict.keys())(**dataset_dict)

        datasets_to_return.append(dataset)

    return datasets_to_return

def update_dataset(dataset_id, name, data_type, val, unit_id, metadata={}, flush=True, **kwargs):
    """
        Update an existing dataset
    """

    if dataset_id is None:
        raise HydraError("Dataset must have an ID to be updated.")

    user_id = kwargs.get('user_id')

    dataset = db.DBSession.query(Dataset).filter(Dataset.id==dataset_id).one()
    #This dataset been seen before, so it may be attached
    #to other scenarios, which may be locked. If they are locked, we must
    #not change their data, so new data must be created for the unlocked scenarios
    locked_scenarios = []
    unlocked_scenarios = []
    for dataset_rs in dataset.resourcescenarios:
        if dataset_rs.scenario.locked == 'Y':
            locked_scenarios.append(dataset_rs)
        else:
            unlocked_scenarios.append(dataset_rs)

    #Are any of these scenarios locked?
    if len(locked_scenarios) > 0:
        #If so, create a new dataset and assign to all unlocked datasets.
        dataset = add_dataset(data_type,
                                val,
                                unit_id,
                                metadata=metadata,
                                name=name,
                                user_id=kwargs['user_id'])
        for unlocked_rs in unlocked_scenarios:
            unlocked_rs.dataset = dataset

    else:

        dataset.type  = data_type
        dataset.value = val
        dataset.set_metadata(metadata)

        dataset.unit_id = unit_id
        dataset.name  = name
        dataset.created_by = kwargs['user_id']
        dataset.hash  = dataset.set_hash()

        #Is there a dataset in the DB already which is identical to the updated dataset?
        existing_dataset = db.DBSession.query(Dataset).filter(Dataset.hash==dataset.hash, Dataset.id != dataset.id).first()
        if existing_dataset is not None and existing_dataset.check_read_permission(user_id, do_raise=False) is True:
            log.warning("An identical dataset %s has been found to dataset %s."
                     " Deleting dataset and returning dataset %s",
                     existing_dataset.id, dataset.id, existing_dataset.id)
            db.DBSession.delete(dataset)
            dataset = existing_dataset
    if flush==True:
        db.DBSession.flush()

    return dataset


def add_dataset(data_type, val, unit_id=None, metadata={}, name="", user_id=None, flush=False):
    """
        Data can exist without scenarios. This is the mechanism whereby
        single pieces of data can be added without doing it through a scenario.

        A typical use of this would be for setting default values on types.
    """

    d = Dataset()

    d.type  = data_type
    d.value = val
    d.set_metadata(metadata)

    d.unit_id  = unit_id
    d.name  = name
    d.created_by = user_id
    d.hash  = d.set_hash()

    try:
        existing_dataset = db.DBSession.query(Dataset).filter(Dataset.hash==d.hash).one()
        if existing_dataset.check_read_permission(user_id, do_raise=False) is True:
            d = existing_dataset
        else:
            d.set_hash()
            db.DBSession.add(d)
    except NoResultFound:
        db.DBSession.add(d)

    if flush == True:
        db.DBSession.flush()
    return d

def bulk_insert_data(data, **kwargs):
    datasets = _bulk_insert_data(data, user_id=kwargs.get('user_id'), source=kwargs.get('app_name'))
    #This line exists to make the db.DBSession 'dirty',
    #thereby telling it to flush the bulk insert.
    datasets[0].name = datasets[0].name

    db.DBSession.flush()

    return datasets

def _make_new_dataset(dataset_dict):
    #If the user is not allowed to use the existing dataset, a new
    #one must be created. This means a unique hash must be created
    #To create a unique hash, add a unique piece of metadata.
    new_dataset = copy.deepcopy(dataset_dict)
    new_dataset['metadata']['created_at'] = datetime.datetime.now()
    new_hash = generate_data_hash(new_dataset)
    new_dataset['hash'] = new_hash

    return new_dataset

def _bulk_insert_data(bulk_data, user_id=None, source=None):
    """
        Insert lots of datasets at once to reduce the number of DB interactions.
        user_id indicates the user adding the data
        source indicates the name of the app adding the data
        both user_id and source are added as metadata
    """
    get_timing = lambda x: datetime.datetime.now() - x
    start_time=datetime.datetime.now()
    new_data = _process_incoming_data(bulk_data, user_id, source)
    log.info("Incoming data processed in %s", (get_timing(start_time)))

    existing_data = _get_existing_data(new_data.keys())

    log.info("Existing data retrieved.")

    #The list of dataset IDS to be returned.
    hash_id_map = {}
    new_datasets = []
    metadata         = {}
    #This is what gets returned.
    for d in bulk_data:

        #limit the name to 200
        d.name = d.name[0:200]

        dataset_dict = new_data[d.hash]
        current_hash = d.hash

        #if this piece of data is already in the DB, then
        #there is no need to insert it!
        if existing_data.get(current_hash) is not None:

            dataset = existing_data.get(current_hash)

            #Is this user allowed to use this dataset?
            #TODO is this too slow?
            if dataset.check_read_permission(user_id, do_raise=False) == False:
                new_dataset = _make_new_dataset(dataset_dict)
                new_datasets.append(new_dataset)
                metadata[new_dataset['hash']] = dataset_dict['metadata']
            else:
                hash_id_map[current_hash] = dataset

        elif current_hash in hash_id_map:
            new_datasets.append(dataset_dict)
        else:
            #set a placeholder for a dataset_id we don't know yet.
            #The placeholder is the hash, which is unique to this object and
            #therefore easily identifiable.
            new_datasets.append(dataset_dict)
            hash_id_map[current_hash] = dataset_dict
            metadata[current_hash] = dataset_dict['metadata']

    log.info("Isolating new data %s", get_timing(start_time))
    #Isolate only the new datasets and insert them
    new_data_for_insert = []
    #keep track of the datasets that are to be inserted to avoid duplicate
    #inserts
    new_data_hashes = []
    for d in new_datasets:
        if d['hash'] not in new_data_hashes:
            d['name'] = d['name'][0:200]
            new_data_for_insert.append(d)
            new_data_hashes.append(d['hash'])

    """
    Identify datasets whose size exceeds the external storage threshold,
    add these to external storage rather than the main db, and replace
    the dataset.value of these with a reference to the external ObjectId.
    Update the metadata to indicate the storage location.
    Note that it isn't necessary to update the hashes calculated above,
    index positions are preserved so these still act as unique identifiers
    for datasets and metadata.
    """
    mongo = MongoStorageAdapter()
    mongo_config = mongo.get_mongo_config()
    threshold_sz = mongo_config["threshold"]
    mongo_location_token = mongo_config["direct_location_token"]
    loc_key = mongo_config["value_location_key"]
    mongo_data = {}

    for idx, ds in enumerate(new_data_for_insert):
        ds_size = len(ds["value"])
        if ds_size > threshold_sz:
            mongo_data[idx] = ds["value"]
            ds_metadata = metadata[ds["hash"]]
            ds_metadata[loc_key] = mongo_location_token

    if mongo_data:
        inserted = mongo.bulk_insert_values(list(mongo_data.values()))
        for idx, key in enumerate(mongo_data):
            new_data_for_insert[key]["value"] = str(inserted.inserted_ids[idx])  # Replace ds.values with _id ref

    if len(new_data_for_insert) > 0:
    	#If we're working with mysql, we have to lock the table..
    	#For sqlite, this is not possible. Hence the try: except
        #try:
        #    db.DBSession.execute("LOCK TABLES tDataset WRITE, tMetadata WRITE")
        #except OperationalError:
        #    pass

        log.info("Inserting new data %s", get_timing(start_time))

        db.DBSession.execute(Dataset.__table__.insert(), new_data_for_insert)
        log.info("New data Inserted %s", get_timing(start_time))

        #try:
        #    db.DBSession.execute("UNLOCK TABLES")
        #except OperationalError:
        #    pass


        new_data = _get_existing_data(new_data_hashes)
        log.info("New data retrieved %s", get_timing(start_time))

        for k, v in new_data.items():
            hash_id_map[k] = v

        _insert_metadata(metadata, hash_id_map)
        log.info("Metadata inserted %s", get_timing(start_time))

    returned_ids = []
    for d in bulk_data:
        returned_ids.append(hash_id_map[d.hash])

    log.info("Done bulk inserting data. %s datasets", len(returned_ids))

    return returned_ids

def _insert_metadata(metadata_hash_dict, dataset_id_hash_dict):
    if metadata_hash_dict is None or len(metadata_hash_dict) == 0:
        return

    metadata_list = []
    for _hash, _metadata_dict in metadata_hash_dict.items():
        for k, v in _metadata_dict.items():
            metadata = {}
            metadata['key']  = str(k)
            metadata['value']  = str(v)
            metadata['dataset_id']      = dataset_id_hash_dict[_hash].id
            metadata_list.append(metadata)

    db.DBSession.execute(Metadata.__table__.insert(), metadata_list)

def _process_incoming_data(data, user_id=None, source=None):

    datasets = {}
    for d in data:
        val = d.parse_value()

        if val is None:
            log.info("Cannot parse data (dataset_id=%s). "
                         "Value not available.",d.id)
            continue

        data_dict = {
            'type': d.type,
            'name': d.name,
            'unit_id': d.unit_id,
            'created_by': user_id,
        }

        data_dict['value'] = val

        if d.metadata is not None:
            if isinstance(d.metadata, dict):
                metadata_dict= d.metadata
            else:
                metadata_dict = json.loads(d.metadata)
        else:
            metadata_dict={}

        metadata_keys = [k.lower() for k in metadata_dict]
        if user_id is not None and 'user_id' not in metadata_keys:
            metadata_dict[u'user_id'] = str(user_id)
        if source is not None and 'source' not in metadata_keys:
            metadata_dict[u'source'] = str(source)

        data_dict['metadata'] = metadata_dict

        d.hash = generate_data_hash(data_dict)

        data_dict['hash'] = d.hash
        datasets[d.hash] = data_dict

    return datasets


def get_metadata(dataset_ids, **kwargs):
    return _get_metadata(dataset_ids)


def _get_metadata(dataset_ids):
    """
        Get all the metadata for a given list of datasets
    """
    metadata = []
    if len(dataset_ids) == 0:
        return []
    if len(dataset_ids) > qry_in_threshold:
        idx = 0
        extent = qry_in_threshold
        while idx < len(dataset_ids):
            log.info("Querying %s metadatas", len(dataset_ids[idx:extent]))
            rs = db.DBSession.query(Metadata).filter(Metadata.dataset_id.in_(dataset_ids[idx:extent])).all()
            metadata.extend(rs)
            idx = idx + qry_in_threshold

            if idx + qry_in_threshold > len(dataset_ids):
                extent = len(dataset_ids)
            else:
                extent = extent +qry_in_threshold
    else:
        metadata_qry = db.DBSession.query(Metadata).filter(Metadata.dataset_id.in_(dataset_ids))
        for m in metadata_qry:
            metadata.append(m)

    return metadata

def _get_existing_data(hashes):

    str_hashes = [str(h) for h in hashes]

    hash_dict = {}

    datasets = []
    if len(str_hashes) > qry_in_threshold:
        idx = 0
        extent =qry_in_threshold
        while idx < len(str_hashes):
            log.info("Querying %s datasets", len(str_hashes[idx:extent]))
            rs = db.DBSession.query(Dataset).filter(Dataset.hash.in_(str_hashes[idx:extent])).all()
            datasets.extend(rs)
            idx = idx + qry_in_threshold

            if idx + qry_in_threshold > len(str_hashes):
                extent = len(str_hashes)
            else:
                extent = extent + qry_in_threshold
    else:
        datasets = db.DBSession.query(Dataset).filter(Dataset.hash.in_(str_hashes))


    for r in datasets:
        hash_dict[r.hash] = r

    log.info("Retrieved %s datasets", len(hash_dict))

    return hash_dict

def _get_datasets(dataset_ids):
    """
        Get all the datasets in a list of dataset IDS. This must be done in chunks of 999,
        as sqlite can only handle 'in' with < 1000 elements.
    """

    dataset_dict = {}

    datasets = []
    if len(dataset_ids) > qry_in_threshold:
        idx = 0
        extent =qry_in_threshold
        while idx < len(dataset_ids):
            log.info("Querying %s datasets", len(dataset_ids[idx:extent]))
            rs = db.DBSession.query(Dataset).filter(Dataset.id.in_(dataset_ids[idx:extent])).all()
            datasets.extend(rs)
            idx = idx + qry_in_threshold

            if idx + qry_in_threshold > len(dataset_ids):
                extent = len(dataset_ids)
            else:
                extent = extent + qry_in_threshold
    else:
        datasets = db.DBSession.query(Dataset).filter(Dataset.id.in_(dataset_ids))


    for r in datasets:
        dataset_dict[r.id] = r

    log.info("Retrieved %s datasets", len(dataset_dict))

    return dataset_dict

def get_all_dataset_collections(**kwargs):
    all_collections = db.DBSession.query(DatasetCollection).all()

    return all_collections

def _get_collection(collection_id):
    """
        Get a dataset collection by ID
        :param collection ID
    """
    try:
        collection = db.DBSession.query(DatasetCollection).filter(DatasetCollection.id==collection_id).one()
        return collection
    except NoResultFound:
        raise ResourceNotFoundError("No dataset collection found with id %s"%collection_id)

def _get_collection_item(collection_id, dataset_id):
    """
        Get a single dataset collection entry by collection ID and dataset ID
        :param collection ID
        :param dataset ID
    """
    collection_item = db.DBSession.query(DatasetCollectionItem).\
            filter(DatasetCollectionItem.collection_id==collection_id,
                   DatasetCollectionItem.dataset_id==dataset_id).first()
    return collection_item

def add_dataset_to_collection(dataset_id, collection_id, **kwargs):
    """
        Add a single dataset to a dataset collection.
    """
    collection_i = _get_collection(collection_id)
    collection_item = _get_collection_item(collection_id, dataset_id)
    if collection_item is not None:
        raise HydraError("Dataset Collection %s already contains dataset %s", collection_id, dataset_id)

    new_item = DatasetCollectionItem()
    new_item.dataset_id=dataset_id
    new_item.collection_id=collection_id

    collection_i.items.append(new_item)

    db.DBSession.flush()

    return 'OK'


def add_datasets_to_collection(dataset_ids, collection_id, **kwargs):
    """
        Add multiple datasets to a dataset collection.
    """
    collection_i = _get_collection(collection_id)

    for dataset_id in dataset_ids:
        collection_item = _get_collection_item(collection_id, dataset_id)
        if collection_item is not None:
            raise HydraError("Dataset Collection %s already contains dataset %s", collection_id, dataset_id)

        new_item = DatasetCollectionItem()
        new_item.dataset_id=dataset_id
        new_item.collection_id=collection_id

        collection_i.items.append(new_item)

    db.DBSession.flush()
    return 'OK'

def remove_dataset_from_collection(dataset_id, collection_id, **kwargs):
    """
        Add a single dataset to a dataset collection.
    """
    collection_i = _get_collection(collection_id)
    collection_item = _get_collection_item(collection_id, dataset_id)
    if collection_item is None:
        raise HydraError("Dataset %s is not in collection %s.",
                                                    dataset_id,
                                                    collection_id)
    db.DBSession.delete(collection_item)
    db.DBSession.flush()

    db.DBSession.expunge_all()

    return 'OK'


def check_dataset_in_collection(dataset_id, collection_id, **kwargs):
    """
        Check whether a dataset is contained inside a collection
        :param dataset ID
        :param collection ID
        :returns 'Y' or 'N'
    """

    _get_collection(collection_id)
    collection_item = _get_collection_item(collection_id, dataset_id)
    if collection_item is None:
        return 'N'
    else:
        return 'Y'



def get_dataset_collection(collection_id,**kwargs):
    try:
        collection = db.DBSession.query(DatasetCollection).filter(DatasetCollection.id==collection_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("No dataset collection found with id %s"%collection_id)
    #lazy load items
    collection.items
    return collection

def get_dataset_collection_by_name(collection_name,**kwargs):
    try:
        collection = db.DBSession.query(DatasetCollection).filter(DatasetCollection.name==collection_name).one()
    except NoResultFound:
        raise ResourceNotFoundError("No dataset collection found with name %s"%collection_name)
    #lazy load items
    collection.items
    return collection

def add_dataset_collection(collection,**kwargs):

    coln_i = DatasetCollection(name=collection.name)

    for dataset_id in collection.dataset_ids:
        datasetitem = DatasetCollectionItem(dataset_id=dataset_id)
        coln_i.items.append(datasetitem)
    db.DBSession.add(coln_i)
    db.DBSession.flush()
    return coln_i

def delete_dataset_collection(collection_id,**kwargs):

    try:
        collection = db.DBSession.query(DatasetCollection).filter(DatasetCollection.id==collection_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("No dataset collection found with id %s"%collection_id)

    db.DBSession.delete(collection)
    db.DBSession.flush()

def get_collections_like_name(collection_name,**kwargs):
    """
        Get all the datasets from the collection with the specified name
    """
    try:
        collections = db.DBSession.query(DatasetCollection).filter(
            DatasetCollection.name.like("%%%s%%"%collection_name.lower())).all()
    except NoResultFound:
        raise ResourceNotFoundError("No dataset collection found with name %s"%collection_name)
    #lazy load items
    [collection.items for collection in collections]
    return collections

def get_collection_datasets(collection_id,**kwargs):
    """
        Get all the datasets from the collection with the specified name
    """
    collection_datasets = db.DBSession.query(Dataset).filter(Dataset.id==DatasetCollectionItem.dataset_id,
                                        DatasetCollectionItem.collection_id==DatasetCollection.id,
                                        DatasetCollection.id==collection_id).all()
    return collection_datasets

def get_val_at_time(dataset_id, timestamps,**kwargs):
    """
    Given a timestamp (or list of timestamps) and some timeseries data,
    return the values appropriate to the requested times.

    If the timestamp is before the start of the timeseries data, return
    None If the timestamp is after the end of the timeseries data, return
    the last value.  """

    t = []
    for time in timestamps:
        t.append(get_datetime(time))
    dataset_i = db.DBSession.query(Dataset).filter(Dataset.id==dataset_id).one()
    #for time in t:
    #    data.append(td.get_val(timestamp=time))

    data = dataset_i.get_val(timestamp=t)
    if data is not None:
        dataset = JSONObject({'data': json.dumps(data)})
    else:
        dataset = JSONObject({'data': None})

    return dataset

def get_multiple_vals_at_time(dataset_ids, timestamps,**kwargs):
    """
    Given a timestamp (or list of timestamps) and a list of timeseries datasets,
    return the values appropriate to the requested times.

    If the timestamp is before the start of the timeseries data, return
    None If the timestamp is after the end of the timeseries data, return
    the last value.  """
    datasets = _get_datasets(dataset_ids)
    datetimes = []
    for time in timestamps:
        datetimes.append(get_datetime(time))

    return_vals = {}
    for dataset_i in datasets.values():
        data = dataset_i.get_val(timestamp=datetimes)
        ret_data = {}
        if type(data) is list:
            for i, t in enumerate(timestamps):
                ret_data[t] = data[i]
        return_vals['dataset_%s'%dataset_i.id] = ret_data

    return return_vals

def get_vals_between_times(dataset_id, start_time, end_time, timestep, increment, **kwargs):
    """
        Retrive data between two specified times within a timeseries. The times
        need not be specified in the timeseries. This function will 'fill in the blanks'.

        Two types of data retrieval can be done.

        If the timeseries is timestamp-based, then start_time and end_time
        must be datetimes and timestep must be specified (minutes, seconds etc).
        'increment' reflects the size of the timestep -- timestep = 'minutes' and increment = 2
        means 'every 2 minutes'.

        If the timeseries is float-based (relative), then start_time and end_time
        must be decimal values. timestep is ignored and 'increment' represents the increment
        to be used between the start and end.
        Ex: start_time = 1, end_time = 5, increment = 1 will get times at 1, 2, 3, 4, 5
    """

    try:
        server_start_time = get_datetime(start_time)
        server_end_time   = get_datetime(end_time)
        times = [server_start_time]

        next_time = server_start_time
        while next_time < server_end_time:
            if int(increment) == 0:
                raise HydraError("%s is not a valid increment for this search."%increment)
            next_time = next_time  + datetime.timedelta(**{timestep:int(increment)})
            times.append(next_time)
    except ValueError:
        try:
            server_start_time = Decimal(start_time)
            server_end_time   = Decimal(end_time)
            times = [float(server_start_time)]

            next_time = server_start_time
            while next_time < server_end_time:
                next_time = float(next_time) + increment
                times.append(next_time)
        except:
            raise HydraError("Unable to get times. Please check to and from times.")

    td = db.DBSession.query(Dataset).filter(Dataset.id==dataset_id).one()
    log.debug("Number of times to fetch: %s", len(times))

    data = td.get_val(timestamp=times)

    data_to_return = []
    if type(data) is list:
        for d in data:
            if d is not None:
                data_to_return.append(list(d))
    elif data is None:
        data_to_return = []
    else:
        data_to_return.append(data)

    dataset = JSONObject({'data' : json.dumps(data_to_return)})

    return dataset

def delete_dataset(dataset_id,**kwargs):
    """
        Removes a piece of data from the DB.
        CAUTION! Use with care, as this cannot be undone easily.
    """
    try:
        d = db.DBSession.query(Dataset).filter(Dataset.id==dataset_id).one()
    except NoResultFound:
        raise HydraError("Dataset %s does not exist."%dataset_id)

    dataset_rs = db.DBSession.query(ResourceScenario).filter(ResourceScenario.dataset_id==dataset_id).all()
    if len(dataset_rs) > 0:
        raise HydraError("Cannot delete %s. Dataset is used by one or more resource scenarios."%dataset_id)

    db.DBSession.delete(d)

    db.DBSession.flush()

    #Remove ORM references to children of this dataset (metadata, collection items)
    db.DBSession.expunge_all()

def read_json(json_string):
    pd.read_json(json_string)

def get_hdf_file_size(url, **kwargs):
    """
      Returns the size in bytes of the hdf file <url> argument
      Raises ValueError on bad url
    """
    hdf = HdfStorageAdapter()
    return hdf.file_size(url)

def get_hdf_series_info(url, groupname=None, columns=None, **kwargs):
    """
      Returns information about the series contained in the
      <columns> argument in the specified <groupname>

      Returns an object containing two keys:
        - index: an object of 'name', 'length', 'dtype' for
                 the index of the <group> arg
        - series: an array of objects, each containing 'name',
                  'length', 'dtype' for each column of the
                  <group> arg

      If <columns> is None or is an empty container (including a
      zero-length string), info on *all* series in the group
      <groupname> will be returned.

      If <columns> is a non-empty string, it is considered to
      represent a single column name.

      Raises ValueError on bad arguments.
    """
    hdf = HdfStorageAdapter()
    return hdf.get_series_info(url, groupname, columns)

def get_hdf_index_info(url, groupname, **kwargs):
    """
      Returns an object with keys...
        {name: str, length: int, dtype: str}
      ...which describes the index of the <groupname> argument.
    """
    hdf = HdfStorageAdapter()
    return hdf.get_index_info(url, groupname)

def get_hdf_index_range(url, groupname, start=0, end=None, **kwargs):
    """
      Returns index entries in the range [start,end) of the <groupname>
      argument.
    """
    hdf = HdfStorageAdapter()
    return hdf.get_index_range(url, groupname, start, end)

def get_hdf_group_info(url, groupname=None, **kwargs):
    """
      Returns an object containing two keys:
        - index: an object of 'name', 'length', 'dtype' for
                 the index of the <group> arg
        - series: an array of objects, each containing 'name',
                  'length', 'dtype' for each column of the
                  <group> arg
    """
    hdf = HdfStorageAdapter()
    return hdf.get_group_info(url, groupname)

def get_hdf_columns_as_dataframe(url, columns, groupname=None, start=None, end=None, **kwargs):
    """
      Returns the rows from <start> to <end> of series specified in
      columns=["col1", "col2", "col3, ...] in the hdf file <url> as
      the json representation of a Pandas DataFrame. This may then
      be read directly in a client with pandas.read_json().

      Keyword arguments:
        <groupname> if absent assume file contains a single group
        <start> start group row, 0 if absent
        <end> end group row, len(groupdata) if absent

      Raises ValueError on bad url, dataset name or bounds
    """
    hdf = HdfStorageAdapter()
    return hdf.get_columns_as_dataframe(url, groupname, columns, start, end)

def resolve_url_to_path(url, **kwargs):
    """
      Returns the path, either on a local filesystem or remote
      storage, to which the <url> arg resolves.
      Raises ValueError if the arg cannot be interpreted as a valid url
    """
    hdf = HdfStorageAdapter()
    return hdf.url_to_filestore_path(url)

def file_exists_at_url(url, **kwargs):
    """
      Return a boolean corresponding to the existence of a readable
      file at the <url> arg.
      Never raises, returns False for invalid args or no permission
    """
    hdf = HdfStorageAdapter()
    return hdf.file_exists_at_url(url)

def get_hdf_group_as_dataframe(url, **kwargs):
    """
      Return the entire table of series within an HDF group as a
      single dataframe.  Timeseries indices are represented in iso8601 format
      Keyword arguments:
        <groupname> if absent assume file contains a single group
        <start> start group row, 0 if absent
        <end> end group row, len(groupdata) if absent
    """
    hdf = HdfStorageAdapter()
    return hdf.get_columns_as_dataframe(url, columns=None, **kwargs)

def get_hdf_groups(url, **kwargs):
    """
      Returns the root groups of an HDF file.
      These may subsequently be used as the <groupname> arg to
      `get_hdf_group_as_dataframe`
    """
    hdf = HdfStorageAdapter()
    return hdf.get_hdf_groups(url)

def get_hdf_group_columns(url, groupname, **kwargs):
    """
      Returns a list containing the names of series within
      the dataset of a specified group
    """
    hdf = HdfStorageAdapter()
    return hdf.get_group_columns(url, groupname)

@required_role("admin")
def retrieve_s3_file_to_local_storage(url, **kwargs):
    """
      Forces retrieval of the s3 file at <url> to Hydra's
      local filesystem.
      Future references to the original url may then be
      replaced with the <file_path> location returned
    """
    hdf = HdfStorageAdapter()
    file_path, file_size = hdf.retrieve_s3_file(url)
    return file_path, file_size

@required_role("admin")
def list_local_files(**kwargs):
    """
      Returns an object of filename: file_size mappings
      describing files in the HDF filestore
    """
    hdf = HdfStorageAdapter()
    return hdf.list_local_files()

@required_role("admin")
def purge_local_file(filename, **kwargs):
    """
      Removes a file from the HDF filestore.
      Returns the path of the deleted filename on
      success.
      Raises ValueError on failure.
    """
    hdf = HdfStorageAdapter()
    return hdf.purge_local_file(filename)



"""network_appdata

Revision ID: cec2b77ad85e
Revises: 1d3ab26415ec
Create Date: 2021-02-12 10:10:35.646470

"""
import logging
from alembic import op
import sqlalchemy as sa

log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'cec2b77ad85e'
down_revision = '1d3ab26415ec'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tNetwork',
                          sa.Column('appdata',
                                    sa.Text().with_variant(sa.dialects.mysql.LONGTEXT, 'mysql')))
        except Exception as e:
            log.critical(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.drop_column('tNetwork', 'appdata')
        except Exception as e:
            log.critical(e)



"""
Utilities to assist with managing datasets in external storage.

It should be noted that when exporting and importing datasets
to and from external storage, the Hydra config size threshold
is ignored, but this is still applied to any changes to datasets
performed via Hydra.
"""
import bz2
import io
import json
import logging
import os
import transaction

from bson.objectid import ObjectId
import pymongo
from pymongo import MongoClient
from sqlalchemy.sql.expression import func
from sqlalchemy.exc import NoResultFound

from hydra_base import db
from hydra_base.db.model import (
    Dataset,
    Metadata
)
from hydra_base.lib.storage import MongoStorageAdapter

from typing import (
    Dict,
    List,
    Tuple
)

log = logging.getLogger(__name__)

if not db.DBSession:
    db.connect()

mongo = None


def largest_datasets(count: int=20) -> List[Tuple[int, int]]:
    """
    Identify the `count` largest datasets in the SQL db and return ids
    and sizes of these.
    """
    datasets = db.DBSession.query(Dataset.id, func.length(Dataset.value))\
                           .order_by(func.length(Dataset.value).desc())\
                           .limit(count).all()
    return datasets


def datasets_larger_than(size: int) -> List[Tuple[int, int]]:
    """
    Identify any datasets larger than `size` chars and return the ids and
    sizes of these.
    """
    datasets = db.DBSession.query(Dataset.id, func.length(Dataset.value))\
                           .filter(func.length(Dataset.value) > size)\
                           .order_by(func.length(Dataset.value).desc()).all()
    return datasets


def total_datasets() -> int:
    """
    Return the total number of datasets in the SQL db
    """
    return db.DBSession.query(Dataset).count()


def dataset_distribution(buckets: int=10) -> Dict:
    """
    Return a histogram of the distribution of dataset sizes in bytes
    consisting of <buckets> divisions
    """
    max_sz = largest_datasets(1)[0][1]
    bucket_sz = max_sz/buckets

    hist = [{"lower": int(i*bucket_sz), "upper": int((i+1)*bucket_sz)} for i in range(buckets)]

    for bucket in hist:
        lower, upper = bucket["lower"], bucket["upper"]
        bucket["count"] = db.DBSession.query(Dataset.id, func.length(Dataset.value))\
                           .filter(func.length(Dataset.value) > lower)\
                           .filter(func.length(Dataset.value) <= upper)\
                           .count()
    return hist


def dataset_report() -> Dict:
    """
    Return a report containing basic statistics describing datasets
    in the SQL database
    """
    total_sz = db.DBSession.query(func.sum(Dataset.value)).scalar()
    mean = db.DBSession.query(func.avg(Dataset.value)).scalar()
    report = {
        "count": total_datasets(),
        "total_size": int(total_sz),
        "mean_size": round(mean, 2),
        "distribution": dataset_distribution()
    }

    return report


def collection_report(db_name: str=None, collection: str=None) -> Dict:
    """
    Return a report containing basic statistics describing documents
    in the specified <collection> of <db_name> Mongo database
    """
    mdb, coll = get_db_and_collection(db_name, collection)
    stats = mdb.command("collstats", coll.name)

    report = {
        "count": stats["count"],
        "total_size": stats["size"],
        "mean_size": stats["avgObjSize"],
        "distribution": document_distribution(db_name, collection)
    }

    return report


def largest_documents(count: int=20, db_name: str=None, collection: str=None) -> List[Tuple[ObjectId, int]]:
    """
    Identifies the <count> largest documents in the specified <collection>
    of the Mongo database <db_name>
    """
    _, coll = get_db_and_collection(db_name, collection)

    pipeline = [
        {"$match": {"value": {"$exists": True}}},
        {"$match": {"value": {"$type": "string"}}},  # Must ensure string args to strLenCP below
        {"$project": {
            "length": {"$strLenCP": "$value"}
            }
        },
        {"$sort": {"length": pymongo.DESCENDING}},
        {"$limit": count}
    ]

    agg = coll.aggregate(pipeline)
    return [(d["_id"], d["length"]) for d in agg]


def document_distribution(db_name: str=None, collection: str=None, buckets: str=10) -> Dict:
    """
    Returns a histogram consisting of <buckets> bins describing the distribution of
    dataset sizes in documents of the <collection> in <db_name>
    """
    _, coll = get_db_and_collection(db_name, collection)

    _, max_sz = largest_documents(1, db_name, collection)[0]
    bucket_sz = max_sz/buckets
    hist = [{"lower": int(i*bucket_sz), "upper": int((i+1)*bucket_sz)} for i in range(buckets)]

    for bucket in hist:
        pipeline = [
            {"$match": {"value": {"$exists": True}}},
            {"$match": {"value": {"$type": "string"}}},  # Must ensure string args to strLenCP below
            {"$redact": {"$cond": [ {"$gt": [{"$strLenCP": "$value"}, bucket["lower"]]}, "$$KEEP", "$$PRUNE"]}},
            {"$redact": {"$cond": [ {"$lte": [{"$strLenCP": "$value"}, bucket["upper"]]}, "$$KEEP", "$$PRUNE"]}},
            {"$count": "count"}
        ]

        result = coll.aggregate(pipeline)
        rl = [*result]
        count = rl[0]["count"] if rl else 0
        bucket["count"] = count

    return hist


def export_dataset_to_external_storage(ds_id: int, db_name: str=None, collection: str=None):
    """
    Place the value of the dataset identified by `ds_id` in external
    storage, replace the value with an ObjectID reference, and
    update the dataset metadata to indicate the new location.
    """
    mongo_config = MongoStorageAdapter.get_mongo_config()
    db_name = db_name if db_name else mongo_config["db_name"]
    collection = collection if collection else mongo_config["datasets"]

    mongo = get_mongo_client()
    path = mongo[db_name][collection]

    dataset = db.DBSession.query(Dataset).filter(Dataset.id == ds_id).one()
    """ Verify dataset does not already have external storage metadata """
    if dataset.is_external():
        raise LookupError(f"Dataset {dataset.id} has external storage metadata")

    result = path.insert_one({"value": dataset.value, "dataset_id": dataset.id})
    if not (hasattr(result, "inserted_id") and isinstance(result.inserted_id, ObjectId)):
        raise TypeError(f"Insertion of dataset {dataset.id} to path {db_name}:{collection} failed")

    dataset.value_ref = str(result.inserted_id)
    location_key = mongo_config["value_location_key"]
    external_token = mongo_config["direct_location_token"]
    md = Metadata(key=location_key, value=external_token)
    dataset.metadata.append(md)
    transaction.commit()

    return result


def import_dataset_from_external_storage(ds_id: int, db_name: str=None, collection: str=None):
    """
    Retrieve the value of the dataset identified by `ds_id` from
    external storage, and replace the SQL db dataset value with this.
    Remove any metadata associated with the external storage location.
    """
    dataset = db.DBSession.query(Dataset).filter(Dataset.id == ds_id).one()
    if not dataset.is_external():
        raise LookupError(f"Dataset {dataset.id} does not have external storage metadata")

    mongo_config = MongoStorageAdapter.get_mongo_config()
    db_name = db_name if db_name else mongo_config["db_name"]
    collection = collection if collection else mongo_config["datasets"]

    mongo = get_mongo_client()
    path = mongo[db_name][collection]

    object_id = ObjectId(dataset.value_ref)
    doc = path.find_one({"_id": object_id})
    if not doc:
        raise LookupError(f"No external document {object_id} found for dataset {ds_id} in {db_name}:{collection}")

    """
    If the doc has a reverse reference to a dataset, ensure
    it refers to the correct dataset_id
    """
    if doc_ds_id := doc.get("dataset_id"):
        if doc_ds_id != ds_id:
            raise LookupError(f"External doc {object_id} referred to by\
                dataset {ds_id} claims to represent dataset {doc_ds_id}")

    dataset.value_ref = doc["value"]

    location_key = mongo_config["value_location_key"]
    for idx, m in enumerate(dataset.metadata):
        if m.key == location_key:
            break
    dataset.metadata.pop(idx)
    transaction.commit()

    result = path.delete_one({"_id": object_id})
    if result.deleted_count != 1:
        warntext = f"Unable to delete document {object_id} from {db_name}:{collection}"
        raise Warning(warntext)

    return result


def write_dataset_as_bz2(dataset_id: int, path: str='.'):
    """
    Makes a compressed copy of dataset with id <dataset_id> in a
    file named "<dataset_id>.bz2". An optional target directory may
    be provided in <path>
    """
    if not os.path.isdir(path):
        raise ValueError(f"Invalid path {path}; must be an existing directory")

    try:
        dataset = db.DBSession.query(Dataset).filter(Dataset.id == dataset_id).one()
    except NoResultFound:
        raise ValueError(f"No dataset found with id {dataset_id}")

    filename = os.path.join(path, f"{dataset.id}.bz2")
    if os.path.exists(filename):
        raise ValueError(f"File {filename} already exists")

    with bz2.BZ2File(filename, 'wb') as outfile:
        with io.TextIOWrapper(outfile, encoding='utf-8') as tiw:
            ds_size = tiw.write(dataset.value)

    f_size = os.stat(filename).st_size
    log.info(f"Written {filename} {ds_size} => {f_size} bytes ({(1-f_size/ds_size)*100.0:.2f}%)")
    return filename, f_size


def write_oid_as_bz2(oid: str, path: str='.', db_name=None, collection=None):
    """
    Makes a compressed copy of the 'value' attribute of a MongoDB document
    with oid <oid> in a file named "<oid>.bz2". An optional target directory
    may be provided in <path>
    """
    if not os.path.isdir(path):
        raise ValueError(f"Invalid path {path}; must be an existing directory")

    filename = os.path.join(path, f"{oid}.bz2")
    if os.path.exists(filename):
        raise ValueError(f"File {filename} already exists")

    mongo_config = MongoStorageAdapter.get_mongo_config()
    db_name = db_name if db_name else mongo_config["db_name"]
    collection = collection if collection else mongo_config["datasets"]

    mongo = get_mongo_client()
    path = mongo[db_name][collection]

    object_id = ObjectId(oid)
    doc = path.find_one({"_id": object_id})
    if not doc:
        raise LookupError(f"No external document {object_id} found in {db_name}:{collection}")

    with bz2.BZ2File(filename, 'wb') as outfile:
        with io.TextIOWrapper(outfile, encoding='utf-8') as tiw:
            ds_size = tiw.write(doc["value"])

    f_size = os.stat(filename).st_size
    log.info(f"Written {filename} {ds_size} => {f_size} bytes ({(1-f_size/ds_size)*100.0:.2f}%)")
    return filename, f_size


def bz2_file_equal_to_dataset(filename: str) -> bool:
    """
    Verifies that the bz2-compressed dataset contained in <filename> is
    equal to the SQL db dataset with the same dataset_id, as determined
    by the filename prefix.
    """
    _, ds_file = os.path.split(filename)
    dataset_id, _ = os.path.splitext(ds_file)

    try:
        dataset = db.DBSession.query(Dataset).filter(Dataset.id == dataset_id).one()
    except NoResultFound:
        raise ValueError(f"No dataset found with id {dataset_id}")

    with bz2.BZ2File(filename, 'rb') as infile:
        with io.TextIOWrapper(infile, encoding='utf-8') as tiw:
            file_dso = json.loads(tiw.read())

    db_dso = json.loads(dataset.value)
    return db_dso == file_dso


def bz2_file_equal_to_oid(filename: str, db_name=None, collection=None) -> bool:
    """
    Verifies that the bz2-compressed dataset contained in <filename> is
    equal to the MongoDB document with the same dataset_id, as determined
    by the filename prefix.
    """
    _, oid_file = os.path.split(filename)
    oid, _ = os.path.splitext(oid_file)

    mongo_config = MongoStorageAdapter.get_mongo_config()
    db_name = db_name if db_name else mongo_config["db_name"]
    collection = collection if collection else mongo_config["datasets"]

    mongo = get_mongo_client()
    path = mongo[db_name][collection]

    object_id = ObjectId(oid)
    doc = path.find_one({"_id": object_id})
    if not doc:
        raise LookupError(f"No external document {object_id} found in {db_name}:{collection}")

    with bz2.BZ2File(filename, 'rb') as infile:
        with io.TextIOWrapper(infile, encoding='utf-8') as tiw:
            file_oid = json.loads(tiw.read())

    db_oid = json.loads(doc["value"])
    return db_oid == file_oid


def get_mongo_client() -> MongoClient:
    global mongo
    if mongo:
        return mongo
    mongo_config = MongoStorageAdapter.get_mongo_config()
    mongo = MongoClient(f"mongodb://{mongo_config['host']}:{mongo_config['port']}")
    return mongo


def get_db_and_collection(db_name: str=None, collection: str=None):
    mongo_config = MongoStorageAdapter.get_mongo_config()
    db_name = db_name if db_name else mongo_config["db_name"]
    collection = collection if collection else mongo_config["datasets"]

    mongo = get_mongo_client()
    mdb = mongo[db_name]
    coll = mdb[collection]

    return mdb, coll



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *

__all__ = ['ProjectOwner', 'NetworkOwner', 'DatasetOwner']

class ProjectOwner(Base, Inspect):
    """
    """

    __tablename__='tProjectOwner'

    user_id = Column(Integer(), ForeignKey('tUser.id'), primary_key=True, nullable=False)
    project_id = Column(Integer(), ForeignKey('tProject.id'), primary_key=True, nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    view = Column(String(1),  nullable=False, default='Y')
    edit = Column(String(1),  nullable=False, default='N')
    share = Column(String(1),  nullable=False, default='N')

    user = relationship('User')
    project = relationship('Project', backref=backref('owners', order_by=user_id, uselist=True, cascade="all, delete-orphan"))

    _parents  = ['tProject', 'tUser']
    _children = []

    @property
    def read(self):
        return self.view

class NetworkOwner(Base, Inspect):
    """
    """

    __tablename__='tNetworkOwner'

    user_id = Column(Integer(), ForeignKey('tUser.id'), primary_key=True, nullable=False)
    network_id = Column(Integer(), ForeignKey('tNetwork.id'), primary_key=True, nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    view = Column(String(1),  nullable=False, default='Y')
    edit = Column(String(1),  nullable=False, default='N')
    share = Column(String(1),  nullable=False, default='N')

    user = relationship('User')
    network = relationship('Network', backref=backref('owners', order_by=user_id, uselist=True, cascade="all, delete-orphan"))

    _parents  = ['tNetwork', 'tUser']
    _children = []


class DatasetOwner(Base, Inspect):
    """
    """

    __tablename__='tDatasetOwner'

    user_id = Column(Integer(), ForeignKey('tUser.id'), primary_key=True, nullable=False)
    dataset_id = Column(Integer(), ForeignKey('tDataset.id'), primary_key=True, nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    view = Column(String(1),  nullable=False, default='Y')
    edit = Column(String(1),  nullable=False, default='N')
    share = Column(String(1),  nullable=False, default='N')

    user = relationship('User')
    dataset = relationship('Dataset', backref=backref('owners', order_by=user_id, uselist=True, cascade="all, delete-orphan"))

    _parents  = ['tDataset', 'tUser']
    _children = []


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import copy
import datetime
import pytest
import datetime
import json

import hydra_base as hb

import logging
log = logging.getLogger(__name__)

class TestNetwork:
    """
        Test for network-based functionality
    """

    def test_add_network_unknown_attribute(self, client, projectmaker):
        project = projectmaker.create('test')
        project_id = project.id

        net_ra_bad_attr_id = dict(
            ref_id = None,
            ref_key = 'NETWORK',
            attr_is_var = 'N',
            attr_id = 999, # an unknown attribute ID~
            id = -1
        )

        network = dict(
            name = 'Network @ %s'%datetime.datetime.now(),
            description = 'Test network with 2 nodes and 1 link',
            project_id = project_id,
            links = [],
            nodes = [],
            layout = {},
            scenarios = [],
            resourcegroups = [],
            projection = None,
            attributes = [net_ra_bad_attr_id],
        )


        with pytest.raises(hb.exceptions.HydraError):
            client.add_network(network)



    def test_get_network_with_template(self, client, network_with_data):
        """

        """
        net = network_with_data
        logging.info("%s nodes before", len(net.nodes))
        #All the nodes are in this template, so return them all
        assert len(net.nodes) == 10
        #The type has only 2 attributes, so these are the only
        #ones which should be returned.
        for n in net.nodes:
            assert len(n.attributes) == 4

        #only 4 of the links in the network have a type, so only these
        #4 should be returned.
        logging.info("%s links before", len(net.links))
        assert len(net.links) == 9
        #of the 4 links returned, ensure the two attributes are on each one.
        for l in net.links:
            if l.types is not None:
                assert len(l.attributes) == 3
            else:
                assert len(l.attributes) == 3
        assert len(net.resourcegroups) == 1

        template_id = net.nodes[0].types[0].template_id

        filtered_net = client.get_network(net.id, template_id=template_id)
        logging.info("%s nodes after",len(filtered_net.nodes))
        #All the nodes are in this template, so return them all
        assert len(filtered_net.nodes) == 10

        assert len(filtered_net.attributes) == 2

        #The type has only 2 attributes, so these are the only
        #ones which should be returned.
        for n in filtered_net.nodes:
            assert len(n.attributes) == 4
        #only 4 of the links in the network have a type, so only these
        #4 should be returned.
        logging.info("%s links after", len(filtered_net.links))
        assert len(filtered_net.links) == 4
        #of the 4 links returned, ensure the two attributes are on each one.
        for l in filtered_net.links:
            assert len(l.attributes) == 3

        assert len(filtered_net.resourcegroups) == 0

        unfiltered_net = client.get_network(net.id)

        assert len(unfiltered_net.attributes) == 4

        filtered_net_with_extra_attributes = client.get_network(
            net.id,
            template_id=template_id,
            include_non_template_attributes=True)

        assert len(filtered_net.attributes) == len(filtered_net_with_extra_attributes.attributes) - 2

    def test_get_resources_of_type(self, client, network_with_data):
        """
            Test for the retrieval of all the resources of a specified
            type within a network.
        """

        net = network_with_data
        link_ids = []
        type_id = None
        for l in net.links:
            if l.types:
                if type_id is None:
                    type_id = l.types[0].id
                link_ids.append(l.id)

        resources_of_type = list(client.get_resources_of_type(net.id, type_id))

        #this returns a tuple of the node, link and group types. Link types are at
        #index 1 of this tuple
        assert len(resources_of_type) == 4
        for r in list(resources_of_type):
            assert r.ref_key == 'LINK'
            assert r.id in link_ids

    def test_get_all_resource_attributes_in_network(self, client, network_with_data):
        """
            Test to retrieve all the resourceattributes relating to the specified
            attribute ID in the network, regardless of what node, link group etc
            they belong to.
        """
        test_attr_id = None
        for a in network_with_data.nodes[0].attributes:
            test_attr_id = a.attr_id
            break


        all_network_resource_attrs = client.get_all_resource_attributes_in_network(
            test_attr_id,
            network_with_data.id
        )

        #Find the attribute that ALL nodes have.
        assert len(all_network_resource_attrs) == len(network_with_data.nodes)


    def test_get_network_1(self, client, networkmaker):
        """
            Test for the potentially likely case of creating a network with two
            scenarios, then querying for the network without data to identify
            the scenarios, then querying for the network with data but in only
            a select few scenarios.
        """
        net = networkmaker.create(map_projection='EPSG:21781')
        scenario_id = net.scenarios[0].id

        clone = client.clone_scenario(scenario_id)
        new_scenario = client.get_scenario(clone.id)

        full_network = client.get_network(new_scenario.network_id, include_data=False)

        for s in full_network.scenarios:
            assert len(s.resourcescenarios) == 0

        scen_ids = [scenario_id]
        partial_network = client.get_network(new_scenario.network_id, include_attributes=True, include_data=True, include_results=True, scenario_ids=scen_ids)

        assert len(partial_network.scenarios) == 1
        assert len(full_network.scenarios) == 2

        for s in partial_network.scenarios:
            assert len(s.resourcescenarios) > 0

        network_with_results = client.get_network(new_scenario.network_id, include_attributes=True, include_data=True, scenario_ids=scen_ids)
        network_no_results = client.get_network(new_scenario.network_id, include_attributes=True, include_data=True, include_results=False, scenario_ids=scen_ids)

        sample_rs= network_with_results.scenarios[0].resourcescenarios[0]
        #there should be one more result in the
        assert len(network_with_results.scenarios[0].resourcescenarios) == len(network_no_results.scenarios[0].resourcescenarios) + 10
        metadata = json.loads(sample_rs.dataset.metadata) if isinstance(sample_rs.dataset.metadata, str) else sample_rs.dataset.metadata
        assert len(metadata) == 0

        network_with_results_and_metadata = client.get_network(new_scenario.network_id, include_attributes=True, include_data=True, scenario_ids=scen_ids, template_id=None, include_non_template_attributes=None, include_metadata=True)

        sample_rs= network_with_results_and_metadata.scenarios[0].resourcescenarios[0]
        metadata = json.loads(sample_rs.dataset.metadata) if isinstance(sample_rs.dataset.metadata, str) else sample_rs.dataset.metadata
        #there should be one more result in the
        assert len(metadata) > 0



        with pytest.raises(hb.exceptions.HydraError):
            client.get_network_by_name(net.project_id, "I am not a network")

        net_by_name = client.get_network_by_name(net.project_id, net.name)
        assert net_by_name.id == full_network.id

        no_net_exists = client.network_exists(net.project_id, "I am not a network")
        assert no_net_exists == 'N'
        net_exists = client.network_exists(net.project_id, net.name)
        assert net_exists == 'Y'
        assert full_network.projection == 'EPSG:21781'

    def test_get_extents(self, client, network_with_data):
        """
        Extents test: Test that the min X, max X, min Y and max Y of a
        network are retrieved correctly.
        """
        net = network_with_data

        extents = client.get_network_extents(net.id)

        assert extents.min_x == 10
        assert extents.max_x == 100
        assert extents.min_y == 9
        assert extents.max_y == 99

    def test_update_network(self, client, network_with_data):

        net = hb.JSONObject(client.get_network(network_with_data.id))

        link_id = net.links[1].id
        old_node_1_id = net.links[1].node_1_id
        old_node_2_id = net.links[1].node_2_id

        net.links[1].node_1_id = net.nodes[-1].id
        net.links[1].node_2_id = net.nodes[-2].id
        net.links[1].layout = {'color':'red'}

        net.nodes[1].layout = {'color':'green'}

        net.description = \
            'A different network for SOAP unit tests.'

        updated_network = client.update_network(net)

        assert net.id == updated_network.id, \
            'network_id has changed on update.'
        assert net.name == updated_network.name, \
            "network_name changed on update."
        assert updated_network.links[1].id == link_id
        assert updated_network.links[1].node_1_id != old_node_1_id
        assert updated_network.links[1].node_1_id == net.nodes[-1].id
        assert updated_network.links[1].layout['color'] == 'red'

        assert updated_network.links[1].node_2_id != old_node_2_id
        assert updated_network.links[1].node_2_id == net.nodes[-2].id

        assert updated_network.nodes[1].layout['color'] == 'green'


############################################################
    def test_add_links(self, client, projectmaker):

        project = projectmaker.create('test')
        network = hb.JSONObject()
        nodes = []
        links = []

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        for i in range(nnodes):
            node = hb.JSONObject()
            node.id = i * -1
            node.name = 'Node ' + str(i)
            node.description = 'Test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            nodes.append(node)

        for i in range(nlinks):
            link = hb.JSONObject()
            link.id = i * -1
            link.name = 'Link ' + str(i)
            link.description = 'Test link ' + str(i)
            link.node_1_id = nodes[i].id
            link.node_2_id = nodes[i + 1].id

            links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'A network for SOAP unit tests.'
        network.nodes = nodes
        network.links = links

        network = client.add_network(network)
        links = []

        link = hb.JSONObject()
        link.id = i * -1
        link.name = 'New Link'
        link.description = 'Test link ' + str(i)
        link.node_1_id = network.nodes[0].id
        link.node_2_id = network.nodes[2].id
        links.append(link)

        link2 = hb.JSONObject()
        link2.id = i * -2
        link2.name = 'New Link_2'
        link2.description = 'Test link ' + str(i)
        link2.node_1_id = network.nodes[0].id
        link2.node_2_id = network.nodes[2].id
        links.append(link2)

        new_links=client.add_links(network.id, links)

        new_network = client.get_network(network.id)

        assert len(network.links)+len(links) == len(new_network.links); "new nodes were not added correctly_2",

############################################################
    def test_add_link(self, client, projectmaker, template):
        project = projectmaker.create('test')
        network = hb.JSONObject()
        nodes = []
        links = []

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        for i in range(nnodes):
            node = hb.JSONObject()
            node.id = i * -1
            node.name = 'Node ' + str(i)
            node.description = 'Test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            nodes.append(node)

        for i in range(nlinks):
            link = hb.JSONObject()
            link.id = i * -1
            link.name = 'Link ' + str(i)
            link.description = 'Test link ' + str(i)
            link.node_1_id = nodes[i].id
            link.node_2_id = nodes[i + 1].id

            links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'A network for SOAP unit tests.'
        network.nodes = nodes
        network.links = links

        network = client.add_network(network)
        network = client.get_network(network.id)

        link = hb.JSONObject()
        link.id = i * -1
        link.name = 'New Link'
        link.description = 'Test link ' + str(i)
        link.node_1_id = network.nodes[0].id
        link.node_2_id = network.nodes[2].id

        tmpl = template

        type_summary_arr = []

        type_summary      = hb.JSONObject()
        type_summary.id   = tmpl.id
        type_summary.name = tmpl.name
        type_summary.id   = tmpl.templatetypes[1].id
        type_summary.name = tmpl.templatetypes[1].name

        type_summary_arr.append(type_summary)

        link.types = type_summary_arr

        new_link = client.add_link(network.id, link)

        link_attr_ids = []
        for resource_attr in new_link.attributes:
            link_attr_ids.append(resource_attr.attr_id)

        for typeattr in tmpl.templatetypes[1].typeattrs:
            assert typeattr.attr_id in link_attr_ids

        new_network = client.get_network(network.id)

        assert len(new_network.links) == len(network.links)+1; "New node was not added correctly"

    def test_add_node(self, client, projectmaker, template):
        project = projectmaker.create('test')
        network = hb.JSONObject()
        nodes = []
        links = []

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        for i in range(nnodes):
            node = hb.JSONObject()
            node.id = i * -1
            node.name = 'node ' + str(i)
            node.description = 'test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            nodes.append(node)

        for i in range(nlinks):
            link = hb.JSONObject()
            link.id = i * -1
            link.name = 'link ' + str(i)
            link.description = 'test link ' + str(i)
            link.node_1_id = nodes[i].id
            link.node_2_id = nodes[i + 1].id

            links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'a network for soap unit tests.'
        network.nodes = nodes
        network.links = links

        network = client.add_network(network)
        network = client.get_network(network.id)

        node = hb.JSONObject()
        new_node_num = nnodes + 1
        node.id = new_node_num * -1
        node.name = 'node ' + str(new_node_num)
        node.description = 'test node ' + str(new_node_num)
        node.x = 100
        node.y = 101


        tmpl = template

        type_summary_arr = []

        type_summary      = hb.JSONObject()
        type_summary.id   = tmpl.id
        type_summary.name = tmpl.name
        type_summary.id   = tmpl.templatetypes[0].id
        type_summary.name = tmpl.templatetypes[0].name

        type_summary_arr.append(type_summary)

        node.types = type_summary_arr

        new_node = client.add_node(network.id, node)

        node_attr_ids = []
        for resource_attr in new_node.attributes:
            node_attr_ids.append(resource_attr.attr_id)

        for typeattr in tmpl.templatetypes[0].typeattrs:
            assert typeattr.attr_id in node_attr_ids

        new_network = client.get_network(network.id)

        assert len(new_network.nodes) == len(network.nodes)+1; "new node was not added correctly"

    ######################################
    def test_add_nodes(self, client, projectmaker):
        """
        Test add new nodes to network
        """

        project = projectmaker.create('test')
        network = hb.JSONObject()
        nodes = []
        links = []

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        for i in range(nnodes):
            node = hb.JSONObject()
            node.id = i * -1
            node.name = 'node ' + str(i)
            node.description = 'test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            nodes.append(node)

        for i in range(nlinks):
            link = hb.JSONObject()
            link.id = i * -1
            link.name = 'link ' + str(i)
            link.description = 'test link ' + str(i)
            link.node_1_id = nodes[i].id
            link.node_2_id = nodes[i + 1].id
            links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'a network for soap unit tests.'
        network.nodes = nodes
        network.links = links

        network = client.add_network(network)
        network = client.get_network(network.id)
        nodes = []

        for i in range (1200):
             node1 = hb.JSONObject()
             new_node_num = nnodes + 1
             node1.id = new_node_num * -1
             node1.name = 'node1_'+str(i)
             node1.description = 'test node ' + str(new_node_num)
             node1.x = 100+i
             node1.y = 101+i
             nodes.append(node1)

        new_nodes=client.add_nodes(network.id, nodes)
        new_network = client.get_network(network.id)

        assert len(network.nodes)+len(nodes) == len(new_network.nodes); "new nodes were not added correctly_2",

    ########################################


    def test_update_node(self, client, network_with_data):
        network = network_with_data

        node_to_update = hb.JSONObject(network.nodes[0])
        node_to_update.name = "Updated Node Name"
        node_to_update.layout      = {'app': ["Unit Test1", "Unit Test2"]}

        new_node = client.update_node(node_to_update)

        new_network = hb.JSONObject(client.get_network(network.id))

        updated_node = None
        for n in new_network.nodes:
            if n.id == node_to_update.id:
                updated_node = n
        assert updated_node.layout is not None
        assert updated_node.layout['app']  == ["Unit Test1", "Unit Test2"]
        assert updated_node.name == "Updated Node Name"

    def test_set_node_status(self, client, network_with_data):
        network = network_with_data

        node_to_delete = network.nodes[0]

        client.set_node_status(node_to_delete.id, 'X')

        new_network = client.get_network(network.id)

        node_ids = []
        for n in new_network.nodes:
            node_ids.append(n.id)
        for l in new_network.links:
            node_ids.append(l.node_1_id)
            node_ids.append(l.node_2_id)
        assert node_to_delete.id not in node_ids

        client.set_node_status(node_to_delete.id, 'A')

        new_network = client.get_network(network.id)

        node_ids = []
        for n in new_network.nodes:
            node_ids.append(n.id)
        link_node_ids = []
        for l in new_network.links:
            link_node_ids.append(l.node_1_id)
            link_node_ids.append(l.node_2_id)
        assert node_to_delete.id in link_node_ids
        assert node_to_delete.id in link_node_ids
        assert node_to_delete.id in node_ids

    def test_update_link(self, client, network_with_data):
        network = network_with_data

        link_to_update = network.links[0]
        link_to_update.name = "Updated link Name"
        link_to_update.layout      = {'app': ["Unit Test1", "Unit Test2"]}

        new_link = client.update_link(link_to_update)

        new_network = client.get_network(network.id)

        updated_link = None
        for l in new_network.links:
            if l.id == link_to_update.id:
                updated_link = l
        assert updated_link.layout is not None

        assert updated_link.layout['app']  == ["Unit Test1", "Unit Test2"]
        assert updated_link.name == "Updated link Name"

    def test_set_link_status(self, client, network_with_data):
        network = network_with_data

        link_to_delete = network.links[0]

        client.set_link_status(link_to_delete.id, 'X')

        new_network = hb.JSONObject(client.get_network(network.id))

        link_ids = []
        for l in new_network.links:
            link_ids.append(l.id)
        assert link_to_delete.id not in link_ids

        client.set_link_status(link_to_delete.id, 'A')
        new_network = client.get_network(network.id)
        link_ids = []
        for l in new_network.links:
            link_ids.append(l.id)
        assert link_to_delete.id in link_ids


    def test_set_network_status(self, client, projectmaker):
        project = projectmaker.create('test')
        network = hb.JSONObject()
        nodes = []
        links = []

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        for i in range(nnodes):
            node = hb.JSONObject()
            node.id = i * -1
            node.name = 'Node ' + str(i)
            node.description = 'Test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            nodes.append(node)

        for i in range(nlinks):
            link = hb.JSONObject()
            link.id = i * -1
            link.name = 'Link ' + str(i)
            link.description = 'Test link ' + str(i)
            link.node_1_id = nodes[i].id
            link.node_2_id = nodes[i + 1].id
            #link = client.add_link(link)

            links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'A network for SOAP unit tests.'
        network.nodes = nodes
        network.links = links

        network = client.add_network(network)

        client.set_network_status(network.id, 'X')

        assert client.get_network(network.id).status == 'X', \
            'Deleting network did not work correctly.'

        client.set_network_status(network.id, 'A')

        assert client.get_network(network.id).status == 'A', \
            'Reactivating network did not work correctly.'

    def test_delete_network(self, client, projectmaker, network_with_data):

        network = client.get_network(network_with_data.id)

        client.delete_network(network_with_data.id, 'N')

        with pytest.raises(hb.exceptions.HydraError):
            client.get_network(network.id)
        with pytest.raises(hb.exceptions.HydraError):
            client.get_node(network.nodes[0].id)

        network_attributes = client.get_resource_attributes('NETWORK', network.id)
        assert len(network_attributes) == 0

    def test_get_node(self, client, network_with_data):
        network = network_with_data
        n = network.nodes[0]
        s = network.scenarios[0]

        node_without_data = client.get_node(n.id)

        for ra in node_without_data.attributes:
            assert not hasattr(ra, 'resourcescenario') or ra.resourcescenario is None

        node_with_data = client.get_node(n.id, s.id)

        attrs_with_data = []
        for ra in node_with_data.attributes:
            if hasattr(ra, 'resourcescenario') and ra.resourcescenario is not None:
                if ra.resourcescenario:
                    assert isinstance(ra.resourcescenario, hb.JSONObject)
                    attrs_with_data.append(ra.id)
        assert len(attrs_with_data) == 4


    def test_get_link(self, client, network_with_data):
        network = network_with_data
        l = network.links[-1]
        s = network.scenarios[0]

        link_without_data = client.get_link(l.id)

        for ra in link_without_data.attributes:
            assert not hasattr(ra, 'resourcescenario') or ra.resourcescenario is None

        link_with_data = client.get_link(l.id, s.id)

        attrs_with_data = []
        for ra in link_with_data.attributes:
            if ra.resourcescenario is not None:
                attrs_with_data.append(ra.id)
        assert len(attrs_with_data) == 2

    def test_cleanup_network(self, client, network_with_data):
        network = network_with_data

        node_to_delete = network.nodes[0]

        link_ids = []
        for l in network.links:
            if l.node_1_id == node_to_delete.id:
                link_ids.append(l.id)
            if l.node_2_id == node_to_delete.id:
                link_ids.append(l.id)

        client.set_node_status(node_to_delete.id, 'X')

        client.clean_up_network(network.id)

        with pytest.raises(hb.exceptions.HydraError):
            client.get_node(node_to_delete.id)

        for l in link_ids:
            with pytest.raises(hb.exceptions.HydraError):
                client.get_link(l)

    def test_validate_topology(self, client, projectmaker):
        project = projectmaker.create('test')
        network = hb.JSONObject({})
        nodes = []
        links = []

        nnodes = 3
        nlinks = 2
        x = [0, 0, 1]
        y = [0, 1, 0]

        for i in range(nnodes):
            node = hb.JSONObject()
            node.id = i * -1
            node.name = 'Node ' + str(i)
            node.description = 'Test node ' + str(i)
            node.x = x[i]
            node.y = y[i]

            nodes.append(node)

        #NOTE: NOT ADDING ENOUGH LINKS!!
        for i in range(nlinks-1):
            link = hb.JSONObject()
            link.id = i * -1
            link.name = 'Link ' + str(i)
            link.description = 'Test link ' + str(i)
            link.node_1_id = nodes[i].id
            link.node_2_id = nodes[i + 1].id

            links.append(link)

        network.project_id = project.id
        network.name = 'Test @ %s'%(datetime.datetime.now())
        network.description = 'A network for SOAP unit tests.'
        network.nodes = nodes
        network.links = links

        network = client.add_network(network)

        result = client.validate_network_topology(network.id)
        assert len(result) == 1#This means orphan nodes are present

    def test_consistency_of_update(self, client, network_with_data):
        """
            Test to ensure that updating a network which has not changed
            does not cause any changes to the network.
            Procedure:
            1 Create a network.
            2 Immediately update the network without changing it.
            3 Check that the original network and the updated network are identical.
        """
        net = network_with_data

        for node in net.nodes:
            assert node.types is not None and  len(node.types) > 0

        original_net = client.get_network(net.id)

        updated_net_summary = client.update_network(net)

        updated_net = client.get_network(updated_net_summary.id)

        for node in updated_net.nodes:
            assert node.types is not None and  len(node.types) > 0

        for attr in original_net.keys():
            a = original_net.get(attr)
            b = updated_net[attr]
            #assert str(a) == str(b)
            if attr == 'scenarios':
                for s0 in original_net.scenarios:
                    for s1 in updated_net.scenarios:
                        if s0.id == s1.id:
                            for rs0 in s0.resourcescenarios:
                                for rs1 in s1.resourcescenarios:
                                    if rs0.resource_attr_id == rs1.resource_attr_id:
                                        #Leave these logging in as they are used to test
                                        #whether dataset updates work correctly
                                        #logging.info("%s vs %s",rs0.value, rs1.value)
                                        #logging.info(rs0.value.value==rs1.value.value)
                                        assert str(rs0.value) == str(rs1.value)
            else:
                if str(a) != str(b):
                    logging.critical("%s vs %s",str(a), str(b))
                assert str(a) == str(b)

    def test_get_attribute_data(self, client, network_with_data):
        net = network_with_data
        s = net.scenarios[0]

        node_ras = []
        for node in net.nodes:
            for ra in node.attributes:
                node_ras.append(ra.id)

        link_ras = []
        for link in net.links:
            for ra in link.attributes:
                link_ras.append(ra.id)

        group_ras = []
        for group in net.resourcegroups:
            for ra in group.attributes:
                group_ras.append(ra.id)


        new_node_ras = client.get_all_node_data(net.id, s.id)
        for ra in new_node_ras:
            assert ra.resourcescenario is not None
            assert ra.id in node_ras


        node_id_filter = [net.nodes[0].id, net.nodes[1].id]
        new_node_ras = client.get_all_node_data(net.id, s.id, node_id_filter)
        for ra in new_node_ras:
            assert ra.resourcescenario is not None
            assert ra.id in node_ras

        new_link_ras = client.get_all_link_data(net.id, s.id)
        for ra in new_link_ras:
            assert ra.resourcescenario is not None
            assert ra.id in link_ras

        link_id_filter = [net.links[0].id, net.links[1].id]
        new_link_ras = client.get_all_link_data(net.id, s.id, link_id_filter)
        for ra in new_link_ras:
            assert ra.resourcescenario is not None
            assert ra.id in link_ras

        new_group_ras = client.get_all_group_data(net.id, s.id)
        for ra in new_group_ras:
            assert ra.resourcescenario is not None
            assert ra.id in group_ras

        group_id_filter = [net.resourcegroups[0].id]
        new_group_ras = client.get_all_group_data(net.id, s.id, group_id_filter)
        for ra in new_group_ras:
            assert ra.resourcescenario is not None
            assert ra.id in group_ras

    def test_get_all_resource_data(self, client, network_with_data):
        net = network_with_data
        s = net.scenarios[0]

        all_ras = []
        for node in net.nodes:
            for ra in node.attributes:
                all_ras.append(ra.id)

        for link in net.links:
            for ra in link.attributes:
                all_ras.append(ra.id)

        for group in net.resourcegroups:
            for ra in group.attributes:
                all_ras.append(ra.id)

        all_resource_data = client.get_all_resource_data(s.id, include_values=False)
        for rd in all_resource_data:
            assert rd.value is None
            assert int(rd.resource_attr_id) in all_ras

        all_resource_data = client.get_all_resource_data(s.id, include_values=True)
        for rd in all_resource_data:
            assert rd.value is not None
            assert int(rd.resource_attr_id) in all_ras


        truncated_resource_data = client.get_all_resource_data(s.id, include_values=True, include_metadata=True, page_start=0, page_end=1)
        assert len(truncated_resource_data) == 1

    def test_get_resource_data(self, client, network_with_data):
        net = network_with_data
        s = net.scenarios[0]
        node = net.nodes[0]

        node_ras = [a.id for a in node.attributes]

        all_resource_data = client.get_resource_data('NODE', node.id, s.id, include_values=False)
        all_node_types = []
        for rd in all_resource_data:
            assert rd.dataset.value is None
            assert int(rd.resource_attr_id) in node_ras
            all_node_types.append(rd.dataset.type)

        all_node_types = list(set(all_node_types))

        for datatype in all_node_types:

            all_resource_data = client.get_resource_data('NODE', node.id,s.id, include_values=True)
            for rd in all_resource_data:
                assert rd.dataset.value is not None
                assert int(rd.resource_attr_id) in node_ras

            all_resource_data = client.get_resource_data('NODE', node.id,s.id,
                                                        exclude_data_types=[datatype],
                                                        include_values=True)
            for rd in all_resource_data:
                assert rd.dataset.type != datatype
                assert rd.dataset.value is not None
                assert int(rd.resource_attr_id) in node_ras

            all_resource_data = client.get_resource_data('NODE', node.id,s.id,
                                                        include_data_types=[datatype],
                                                        include_values=True)
            for rd in all_resource_data:
                assert rd.dataset.type == datatype
                assert rd.dataset.value is not None
                assert int(rd.resource_attr_id) in node_ras

            all_resource_data = client.get_resource_data('NODE', node.id,s.id,
                                                        include_data_type_values=[datatype],
                                                        include_values=True)
            for rd in all_resource_data:
                if rd.dataset.type == datatype:
                    assert rd.dataset.value is not None
                else:
                    assert rd.dataset.value is None
                assert int(rd.resource_attr_id) in node_ras

            all_resource_data = client.get_resource_data('NODE', node.id,s.id,
                                                        exclude_data_type_values=[datatype],
                                                        include_values=True)
            for rd in all_resource_data:
                if rd.dataset.type == datatype:
                    # there should be no values returned with this type
                    assert rd.dataset.value is None
                else:
                    assert rd.dataset.value is not None
                assert int(rd.resource_attr_id) in node_ras

    def test_delete_node(self, client, network_with_data):
        net = network_with_data
        scenario_id = net.scenarios[0].id
        client.clone_scenario(scenario_id)

        node_id_to_delete = net.nodes[0].id

        node_datasets = client.get_resource_data('NODE', node_id_to_delete, scenario_id)
        log.info("Deleting node %s", node_id_to_delete)
        client.delete_node(node_id_to_delete, 'Y')

        updated_net = client.get_network(net.id, True)

        remaining_node_ids = [n.id for n in updated_net.nodes]

        assert node_id_to_delete not in remaining_node_ids

        for l in updated_net.links:
            assert l.node_1_id != node_id_to_delete
            assert l.node_2_id != node_id_to_delete

        for rs in node_datasets:
            #In these tests, all timeseries are unique to their resources,
            #so after removing the node no timeseries to which it was attached
            #should still exist.
            d = rs.dataset
            if d.type == 'timeseries':
                with pytest.raises(hb.exceptions.HydraError):
                    dataset = client.get_dataset(d.id)

    def test_delete_link(self, client, network_with_data):
        net = network_with_data
        scenario_id = net.scenarios[0].id
        link_id_to_delete = net.links[0].id

        link_datasets = client.get_resource_data('LINK', link_id_to_delete, scenario_id)
        log.info("Deleting link %s", link_id_to_delete)
        client.delete_link(link_id_to_delete, 'Y')

        updated_net = client.get_network(net.id, True)

        remaining_link_ids = [n.id for n in updated_net.links]

        assert link_id_to_delete not in remaining_link_ids

        with pytest.raises(hb.exceptions.HydraError):
            client.get_link(link_id_to_delete)

        for rs in link_datasets:
            #In these tests, all timeseries are unique to their resources,
            #so after removing the link no timeseries to which it was attached
            #should still exist.
            d = rs.dataset
            if d.type == 'timeseries':
                with pytest.raises(hb.exceptions.HydraError):
                    client.get_dataset(d.id)

    def test_get_all_network_owners(self, client, projectmaker, networkmaker):
        proj = projectmaker.create()

        net1 = networkmaker.create(project_id=proj.id)
        client.share_network(net1.id,
                                 ["UserB"],#Not an admin
                                 'N',
                                 'Y')

        networkowners = client.get_all_network_owners()

        assert len(networkowners) >= 2
        networkowners = client.get_all_network_owners([net1.id])
        assert len(networkowners) == 2

        client.login('UserC', 'password')
        #fails because user C is not an admin
        with pytest.raises(hb.exceptions.HydraError):
            networkowners = client.get_all_network_owners([net1.id])
        #log back in as root
        client.login('root', '')

    def test_bulk_set_network_owners(self, client, networkmaker):
        net = networkmaker.create()

        networkowners = client.get_all_network_owners([net.id])

        assert len(networkowners) == 1

        new_owner = hb.JSONObject(dict(
            network_id=net.id,
            user_id=2,
            view='Y',
            edit='Y',
            share='Y',
        ))
        client.bulk_set_network_owners([new_owner])

        networkowners = client.get_all_network_owners([net.id])

        assert len(networkowners) == 2

    def test_clone_network_into_existing_project(self, client, network_with_data):
        net = network_with_data

        recipient_user = client.get_user_by_name('UserA')

        cloned_network_id = client.clone_network(net.id,
                                          recipient_user_id=recipient_user.id,
                                          new_network_name=None,
                                          project_id=None,
                                          project_name=None,
                                          new_project=False)

        cloned_network = client.get_network(cloned_network_id, include_data=True)

        assert cloned_network.name == net.name + " (1)"
        assert len(net.nodes) == len(cloned_network.nodes)
        assert len(net.links) == len(cloned_network.links)
        assert len(net.resourcegroups) == len(cloned_network.resourcegroups)
        assert len(net.scenarios) == len(cloned_network.scenarios)
        assert len(net.scenarios[0].resourcescenarios) == len(cloned_network.scenarios[0].resourcescenarios) + 10 #this ignores results
        assert len(net.scenarios[0].resourcegroupitems) == len(cloned_network.scenarios[0].resourcegroupitems)


        cloned_network_id = client.clone_network(net.id,
                                          recipient_user_id=recipient_user.id,
                                          new_network_name='My New Name',
                                          project_id=None,
                                          project_name=None,
                                          new_project=False)

        cloned_network = client.get_network(cloned_network_id, include_data=True)
        assert cloned_network.name == 'My New Name'

    def test_clone_network_with_scoped_attributes(self, client, network_with_data):
        net = network_with_data

        network_scoped_attributes = client.get_attributes(network_id=net.id)
        assert len(network_scoped_attributes) == 1
        network_scoped_resource_attributes = list(filter(lambda x: x.attr_id==network_scoped_attributes[0].id, net.attributes))
        assert len(network_scoped_resource_attributes) == 1

        recipient_user = client.get_user_by_name('UserA')

        cloned_network_id = client.clone_network(net.id,
                                          recipient_user_id=recipient_user.id,
                                          new_network_name=None,
                                          project_id=None,
                                          project_name=None,
                                          new_project=False)

        cloned_network = client.get_network(cloned_network_id, include_data=True)

        assert net.project_id == cloned_network.project_id

        network_scoped_attributes = client.get_attributes(network_id=cloned_network.id)
        assert len(network_scoped_attributes) == 0

        project_scoped_attributes = client.get_attributes(project_id=cloned_network.project_id)
        #the attribute has been re-scoped to the project, so now there are 2 on the project. one
        #is the project's original scoped attribute, and the other is the one which has been rescoped.
        assert len(project_scoped_attributes) == 2


    def test_clone_network_into_new_project(self, client, network_with_data):
        net = network_with_data

        recipient_user = client.get_user_by_name('UserA')

        cloned_network_id = client.clone_network(net.id,
                                          recipient_user_id=recipient_user.id,
                                          new_network_name=None,
                                          project_id=None,
                                          project_name=None,
                                          new_project=True)

        cloned_network = client.get_network(cloned_network_id, include_data=True)
        #No need to assert that the clone itself worked, as the other test does that.
        assert cloned_network.project_id != net.project_id

    def test_clone_network_with_scoped_attributes_to_new_project(self, client, network_with_data):
        net = network_with_data

        network_scoped_attributes = client.get_attributes(network_id=net.id)
        assert len(network_scoped_attributes) == 1
        network_scoped_resource_attributes = list(filter(lambda x: x.attr_id==network_scoped_attributes[0].id, net.attributes))
        assert len(network_scoped_resource_attributes) == 1

        recipient_user = client.get_user_by_name('UserA')

        cloned_network_id = client.clone_network(net.id,
                                          recipient_user_id=recipient_user.id,
                                          new_network_name=None,
                                          project_id=None,
                                          project_name=None,
                                          new_project=True)

        cloned_network = client.get_network(cloned_network_id, include_data=True)

        assert net.project_id != cloned_network.project_id

        network_scoped_attributes = client.get_attributes(network_id=cloned_network.id)
        assert len(network_scoped_attributes) == 1

        project_scoped_attributes = client.get_attributes(project_id=cloned_network.project_id)
        #the attribute has not been re-scoped to the project, so the project should have no attributes.
        #this project (as it was created outside the project test suite) does not have a default scoped
        #attributes like the projectes created using the ProjectMaker, hence it will have 0
        assert len(project_scoped_attributes) == 0


    def test_clone_node(self, client, network_with_data):

        node_to_clone = network_with_data.nodes[0]

        cloned_node_id = client.clone_node(node_to_clone.id)

        cloned_node = client.get_node(cloned_node_id)

        assert cloned_node.name == f"{node_to_clone.name} (1)"

        assert len(cloned_node.attributes) == len(node_to_clone.attributes)

        scenario = client.get_scenario(network_with_data.scenarios[0].id)
        original_node_data = list(filter(lambda x: x.resource_attr_id in [a.id for a in node_to_clone.attributes],
                                         scenario.resourcescenarios))

        cloned_node_data = list(filter(lambda x: x.resource_attr_id in [a.id for a in cloned_node.attributes],
                                       scenario.resourcescenarios))

        assert len(cloned_node_data) == len(original_node_data)-1 #has no outputs, so has one less dataset


        cloned_node_id_2 = client.clone_node(node_to_clone.id, include_outputs=True)

        cloned_node_2 = client.get_node(cloned_node_id_2)

        assert cloned_node_2.name == f"{node_to_clone.name} (2)"

        assert len(cloned_node_2.attributes) == len(node_to_clone.attributes)

        scenario = client.get_scenario(network_with_data.scenarios[0].id)
        original_node_data = list(filter(lambda x: x.resource_attr_id in [a.id for a in node_to_clone.attributes],
                                         scenario.resourcescenarios))

        cloned_node_data = list(filter(lambda x: x.resource_attr_id in [a.id for a in cloned_node_2.attributes],
                                       scenario.resourcescenarios))

        assert len(cloned_node_data) == len(original_node_data)

        with pytest.raises(hb.exceptions.HydraError):
            cloned_node_id_3 = client.clone_node(node_to_clone.id, name=network_with_data.nodes[1].name)

        name = "Cloned node"
        x = 100
        y = -100
        cloned_node_id_3 = client.clone_node(node_to_clone.id, name=name, new_x=x, new_y=y)

        cloned_node_3 = client.get_node(cloned_node_id_3)

        assert cloned_node_3.name == name
        assert cloned_node_3.x == x
        assert cloned_node_3.y == y



    def test_clone_nodes(self, client, network_with_data):

        nodes_to_clone = network_with_data.nodes[0:1]

        cloned_node_ids = client.clone_nodes([n.id for n in nodes_to_clone])

        for i, cloned_node_id in enumerate(cloned_node_ids):
            cloned_node = client.get_node(cloned_node_id)
            node_to_clone = nodes_to_clone[i]

            assert cloned_node.name == f"{node_to_clone.name} (1)"

            assert len(cloned_node.attributes) == len(node_to_clone.attributes)

            scenario = client.get_scenario(network_with_data.scenarios[0].id)
            original_node_data = list(filter(lambda x: x.resource_attr_id in [a.id for a in node_to_clone.attributes],
                                            scenario.resourcescenarios))

            cloned_node_data = list(filter(lambda x: x.resource_attr_id in [a.id for a in cloned_node.attributes],
                                        scenario.resourcescenarios))

            assert len(cloned_node_data) == len(original_node_data)-1 #has no outputs, so has one less dataset


    def test_cloned_node_name_similarity(self, client, network_with_data):

        node_to_clone = network_with_data.nodes[0]

        updated_similar_name = node_to_clone.name + ' extratext'

        update_node = network_with_data.nodes[1]
        update_node.name = updated_similar_name

        client.update_node(update_node)

        cloned_node_ids = client.clone_nodes([node_to_clone.id])

        changed_name_node = client.get_node(update_node.id)

        assert changed_name_node.name ==  updated_similar_name

        cloned_node = client.get_node(cloned_node_ids[0])

        assert cloned_node.name == f"{node_to_clone.name} (1)"

        second_cloned_node_ids = client.clone_nodes([node_to_clone.id])

        second_cloned_node = client.get_node(second_cloned_node_ids[0])

        assert second_cloned_node.name == f"{node_to_clone.name} (2)"

        third_cloned_node_ids = client.clone_nodes([node_to_clone.id])

        third_cloned_node = client.get_node(third_cloned_node_ids[0])

        assert third_cloned_node.name == f"{node_to_clone.name} (3)"



"""note

Revision ID: d0c45fdc4fac
Revises: 56ea2e814437
Create Date: 2018-04-17 16:28:03.816876

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql


import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'd0c45fdc4fac'
down_revision = '56ea2e814437'
branch_labels = None
depends_on = None

def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tNote

        try:
            op.alter_column('tNote', 'note_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tNote', 'note_text', new_column_name='value', existing_type=sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'), nullable=True)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tNote
        try:
            op.drop_table('tNote_new')
        except:
            log.info("tNote_new isn't there")

        try:
            # ## tNote
            op.create_table(
                'tNote_new',
                    sa.Column('id',sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('ref_key',sa.String(60),  nullable=False, index=True),
                    sa.Column('value',sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'),  nullable=True),
                    sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                    sa.Column('cr_date',sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.Column('scenario_id',sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=False),
                    sa.Column('project_id',sa.Integer(),  sa.ForeignKey('tProject.id'), index=True, nullable=True),
                    sa.Column('network_id',sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True),
                    sa.Column('node_id',sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                    sa.Column('link_id',sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                    sa.Column('group_id',sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),
            )

            op.execute("insert into tNote_new (id, ref_key, value, created_by, cr_date, scenario_id, project_id, network_id, node_id, link_id, group_id) select note_id, ref_key, note_text, created_by, cr_date, scenario_id, project_id, network_id, node_id, link_id, group_id from tNote")

            op.rename_table('tNote','tNote_old')
            op.rename_table('tNote_new', 'tNote')
            op.drop_table('tNote_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tNote

        try:
            op.alter_column('tNote', 'id', new_column_name='note_id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tNote', 'value', new_column_name='note_text', existing_type=sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'), nullable=True)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tNote
        try:
            op.drop_table('tNote_new')
        except:
            log.info("tNote_new isn't there")

        try:
            # ## tNote
            op.create_table(
                'tNote_new',
                    sa.Column('note_id',sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('ref_key',sa.String(60),  nullable=False, index=True),
                    sa.Column('note_text',sa.LargeBinary(),  nullable=True),
                    sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                    sa.Column('cr_date',sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                    sa.Column('scenario_id',sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=False),
                    sa.Column('project_id',sa.Integer(),  sa.ForeignKey('tProject.id'), index=True, nullable=True),
                    sa.Column('network_id',sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True),
                    sa.Column('node_id',sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                    sa.Column('link_id',sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                    sa.Column('group_id',sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),
            )

            op.execute("insert into tNote_new (note_id, ref_key, note_text, created_by, cr_date, scenario_id, project_id, network_id, node_id, link_id, group_id) select id, ref_key, value, created_by, cr_date, scenario_id, project_id, network_id, node_id, link_id, group_id from tNote")

            op.rename_table('tNote','tNote_old')
            op.rename_table('tNote_new', 'tNote')
            op.drop_table('tNote_old')

        except Exception as e:
            log.exception(e)



"""Attribute groups

Revision ID: ffe998929da4
Revises: 1eb51c6159ac
Create Date: 2018-03-16 09:13:30.493034

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql
import logging

# revision identifiers, used by Alembic.
revision = 'ffe998929da4'
down_revision = '1eb51c6159ac'
branch_labels = None
depends_on = None


def upgrade():
    try:
        # ### commands auto generated by Alembic - please adjust! ###
        op.create_table('tAttrGroup',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=200), nullable=False),
        sa.Column('description', sa.Text().with_variant(mysql.TEXT(1000), 'mysql'), nullable=True),
        sa.Column('layout', sa.Text().with_variant(mysql.TEXT(5000), 'mysql'), nullable=True),
        sa.Column('exclusive', sa.String(length=1), server_default=sa.text(u"'N'"), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('cr_date', sa.TIMESTAMP(), server_default=sa.text(u'CURRENT_TIMESTAMP'), nullable=False),
        sa.ForeignKeyConstraint(['project_id'], ['tProject.project_id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'project_id', name='unique attr group name')
        )
        op.create_index(op.f('ix_tAttrGroup_id'), 'tAttrGroup', ['id'], unique=False)
        op.create_table('tAttrGroupItem',
            sa.Column('group_id', sa.Integer(), nullable=False),
            sa.Column('attr_id', sa.Integer(), nullable=False),
            sa.Column('network_id', sa.Integer(), nullable=False),
            sa.ForeignKeyConstraint(['attr_id'], ['tAttr.attr_id'], ),
            sa.ForeignKeyConstraint(['group_id'], ['tAttrGroup.id'], ),
            sa.ForeignKeyConstraint(['network_id'], ['tNetwork.id'], ),
            sa.PrimaryKeyConstraint('group_id', 'attr_id', 'network_id')
        )
    except Exception as e:
        print("Unable to add attr group tables: %s"%(str(e)))

    try:
        op.create_foreign_key(None, 'tLink', 'tNetwork', ['network_id'], ['id'])
        op.create_unique_constraint('unique net name', 'tNetwork', ['name', 'project_id'])
        op.create_foreign_key(None, 'tNetworkOwner', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tNode', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tNote', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tResourceAttr', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tResourceAttrMap', 'tNetwork', ['network_a_id'], ['id'])
        op.create_foreign_key(None, 'tResourceAttrMap', 'tNetwork', ['network_b_id'], ['id'])
        op.create_foreign_key(None, 'tResourceGroup', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tResourceType', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tRule', 'tNetwork', ['network_id'], ['id'])
        op.create_foreign_key(None, 'tScenario', 'tNetwork', ['network_id'], ['id'])
    except Exception as e:
        print ("Unable to set foreign keys: %s" % str(e))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'tScenario', type_='foreignkey')
    op.drop_constraint(None, 'tRule', type_='foreignkey')
    op.drop_constraint(None, 'tResourceType', type_='foreignkey')
    op.drop_constraint(None, 'tResourceGroup', type_='foreignkey')
    op.drop_constraint(None, 'tResourceAttrMap', type_='foreignkey')
    op.drop_constraint(None, 'tResourceAttrMap', type_='foreignkey')
    op.drop_constraint(None, 'tResourceAttr', type_='foreignkey')
    op.drop_constraint(None, 'tNote', type_='foreignkey')
    op.drop_constraint(None, 'tNode', type_='foreignkey')
    op.drop_constraint(None, 'tNetworkOwner', type_='foreignkey')
    op.drop_constraint('unique net name', 'tNetwork', type_='unique')
    op.drop_constraint(None, 'tLink', type_='foreignkey')
    op.drop_table('tAttrGroupItem')
    op.drop_index(op.f('ix_tAttrGroup_id'), table_name='tAttrGroup')
    op.drop_table('tAttrGroup')
    # ### end Alembic commands ###



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import json
import datetime
import logging

import hydra_base
from hydra_base.util.hydra_dateutil import get_datetime
from hydra_base.lib.objects import JSONObject, Dataset
from hydra_base import config

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class TestUtil:
    def __init__(self, client):
        self.client = client
        #avoid calling the 'get unit function a million times'
        self.unitmap = {}

    def get_unit(self, abbreviation):
        """
            Get the ID of a unit from its abbreviation
        """
        if self.unitmap.get(abbreviation) is None:
            unit = self.client.get_unit_by_abbreviation(abbreviation)
            self.unitmap[abbreviation] = unit
            return unit

        return self.unitmap[abbreviation]

    def create_user(self, name, role='admin'):

        existing_user = self.client.get_user_by_name(name)
        if existing_user not in (None, {}):
            return existing_user

        user = JSONObject(dict(
            username=name,
            password="password",
            display_name="test user",
        ))

        new_user = JSONObject(self.client.add_user(user))

        role = JSONObject(self.client.get_role_by_code(role))

        self.client.set_user_role(new_user_id=new_user.id, role_id=role.id)

        return new_user

    def update_template(self, template_id):

        template = JSONObject(self.client.get_template(template_id))
        new_net_attr = self.create_attribute("net_attr_d", dimension='Monetary Value')

        for tmpltype in template.templatetypes:
            if tmpltype.resource_type == 'NETWORK':
                typeattr_1 = JSONObject(dict(
                    attr_id = new_net_attr.id,
                    data_restriction = {'LESSTHAN': 10, 'NUMPLACES': 1},
                    unit_id = self.get_unit('USD').id,
                ))
                tmpltype.typeattrs.append(typeattr_1)
                break

        template = self.client.update_template(template)


    def create_template(self):

        net_attr1 = self.create_attribute("net_attr_a", dimension='Volume')
        net_attr2 = self.create_attribute("net_attr_c", dimension=None)
        link_attr_1 = self.create_attribute("link_attr_a", dimension='Pressure')
        link_attr_2 = self.create_attribute("link_attr_b", dimension='Speed')
        link_attr_3 = self.create_attribute("link_attr_c", dimension='Length')
        node_attr_1 = self.create_attribute("node_attr_a", dimension='Volume')
        node_attr_2 = self.create_attribute("node_attr_b", dimension='Speed')
        node_attr_3 = self.create_attribute("node_attr_c", dimension='Monetary value')
        node_attr_4 = self.create_attribute("node_attr_d", dimension='Volumetric flow rate')
        group_attr_1 = self.create_attribute("grp_attr_1", dimension='Monetary value')
        group_attr_2 = self.create_attribute("grp_attr_2", dimension=None)

        template = JSONObject()
        template['name'] = 'Default Template ' + str(datetime.datetime.now())


        types = []
        #**********************
        #network type         #
        #**********************
        net_type = JSONObject()
        net_type.name = "Default Network"
        net_type.alias = "Test type alias"
        net_type.resource_type = 'NETWORK'

        typeattrs = []

        typeattr_1 = JSONObject()
        typeattr_1.attr_id = net_attr1.id
        typeattr_1.data_restriction = {'LESSTHAN': 10, 'NUMPLACES': 1}
        typeattr_1.unit_id = self.get_unit('m^3').id
        typeattrs.append(typeattr_1)

        typeattr_2 = JSONObject()
        typeattr_2.attr_id = net_attr2.id
        typeattrs.append(typeattr_2)

        net_type.typeattrs = typeattrs

        types.append(net_type)
        #**********************
        # node type           #
        #**********************
        node_type = JSONObject()
        node_type.name = "Default Node"
        node_type.alias = "Test type alias"
        node_type.resource_type = 'NODE'

        typeattrs = []

        typeattr_1 = JSONObject()
        typeattr_1.attr_id = node_attr_1.id
        typeattr_1.data_restriction = {'LESSTHAN': 10, 'NUMPLACES': 1}
        typeattr_1.unit_id = self.get_unit('m^3').id
        typeattrs.append(typeattr_1)

        typeattr_2 = JSONObject()
        typeattr_2.attr_id = node_attr_2.id
        typeattr_2.data_restriction = {'INCREASING': None}
        typeattrs.append(typeattr_2)

        typeattr_3 = JSONObject()
        typeattr_3.attr_id = node_attr_3.id
        typeattrs.append(typeattr_3)

        typeattr_4 = JSONObject()
        typeattr_4.attr_id = node_attr_4.id
        typeattr_4.unit_id = self.get_unit("m^3 s^-1").id
        typeattrs.append(typeattr_4)

        node_type.typeattrs = typeattrs

        types.append(node_type)
        #**********************
        #link type            #
        #**********************
        link_type = JSONObject()
        link_type.name = "Default Link"
        link_type.resource_type = 'LINK'

        typeattrs = []

        typeattr_1 = JSONObject()
        typeattr_1.attr_id = link_attr_1.id
        typeattrs.append(typeattr_1)

        typeattr_2 = JSONObject()
        typeattr_2.attr_id = link_attr_2.id
        typeattrs.append(typeattr_2)

        typeattr_3 = JSONObject()
        typeattr_3.attr_id = link_attr_3.id
        typeattrs.append(typeattr_3)

        link_type.typeattrs = typeattrs

        types.append(link_type)

        #**********************
        #group type           #
        #**********************
        group_type = JSONObject()
        group_type.name = "Default Group"
        group_type.resource_type = 'GROUP'

        typeattrs = []

        typeattr_1 = JSONObject()
        typeattr_1.attr_id = group_attr_1.id
        typeattrs.append(typeattr_1)

        typeattr_2 = JSONObject()
        typeattr_2.attr_id = group_attr_2.id
        typeattrs.append(typeattr_2)

        group_type.typeattrs = typeattrs

        types.append(group_type)

        template.templatetypes = types

        new_template_i = self.client.add_template(template)
        new_template = JSONObject(new_template_i)

        assert new_template.name == template.name, "Names are not the same!"
        assert new_template.id is not None, "New Template has no ID!"
        assert new_template.id > 0, "New Template has incorrect ID!"

        assert len(new_template.templatetypes) == len(types), "Resource types did not add correctly"
        for t in new_template.templatetypes[1].typeattrs:
            assert t.attr_id in (node_attr_1.id, node_attr_2.id, node_attr_3.id, node_attr_4.id),\
                "Node types were not added correctly!"

        for t in new_template.templatetypes[2].typeattrs:
            assert t.attr_id in (link_attr_1.id, link_attr_2.id, link_attr_3.id),\
                "Link types were not added correctly!"

        return new_template

    def create_templatetype(self, template_id):
        """
            Create a template type object (but don't add it to the DB)
            Args:
                template_id: The ID of the template to which to add this type
            Returns:
                JSONObject with some default values and the correct template ID
        """
        attr_1 = self.create_attribute("link_attr_1", dimension='Pressure')
        attr_2 = self.create_attribute("link_attr_2", dimension='Speed')
        attr_3 = self.create_attribute("link_attr_3", dimension='Volume')

        templatetype = JSONObject()
        templatetype.name = "Test type name @ %s"%(datetime.datetime.now())
        templatetype.alias = "%s alias" % templatetype.name
        templatetype.resource_type = 'LINK'
        templatetype.template_id = template_id
        templatetype.layout = {"color": "red", "shapefile": "blah.shp"}

        templatetype.typeattrs = []

        tattr_1 = JSONObject()
        tattr_1.attr_id = attr_1.id
        tattr_1.description = "added type description 1"
        tattr_1.properties = {"add_type_test_property": "property value"}
        templatetype.typeattrs.append(tattr_1)

        tattr_2 = JSONObject()
        tattr_2.attr_id = attr_2.id
        tattr_1.description = "added type description 2"
        templatetype.typeattrs.append(tattr_2)

        tattr_3 = JSONObject()
        tattr_3.attr_id = attr_3.id
        templatetype.typeattrs.append(tattr_3)

        return templatetype

    def create_child_template(self, parent_id):
        """
            Make a template which is a child of a parent template.
        """

        net_attr_child = self.create_attribute("net_attr_child", dimension=None)
        link_attr_child = self.create_attribute("link_attr_child", dimension='Length')
        node_attr_child = self.create_attribute("node_attr_child", dimension='Volumetric flow rate')
        group_attr_child = self.create_attribute("grp_attr_child", dimension='Volume')

        name = 'Child Template ' + str(datetime.datetime.now())

        child_template_j = self.client.add_child_template(parent_id, name=name)

        return child_template_j

    def create_project(self, name=None, share=True, parent_id=None):

        if name is None:
            name = "Unittest Project"

        user_projects = self.client.get_project_by_name(name)

        if len(user_projects) == 0:
            project = JSONObject()
            project.name = name
            project.description = "Project which contains all unit test networks"
            project.parent_id = parent_id
            project = JSONObject(self.client.add_project(project))
            if share is True:
                self.client.share_project(
                    project.id,
                    ["UserA", "UserB", "UserC"], 'N', 'Y')

            proj_scoped_attr = self.create_attribute("Project Scoped Attr", dimension=None, project_id=project.id)


            return project
        else:
            return user_projects[0]


    def create_link(self, link_id, node_1_name, node_2_name, node_1_id, node_2_id):

        ra_array = []

        link = JSONObject({
            'id'          : link_id,
            'name'        : "%s_to_%s"%(node_1_name, node_2_name),
            'description' : 'A test link between two nodes.',
            'layout'      : None,
            'node_1_id'   : node_1_id,
            'node_2_id'   : node_2_id,
            'attributes'  : ra_array,
        })

        return link

    def create_node(self, node_id, attributes=None, node_name="Test Node Name"):

        if attributes is None:
            attributes = []
        #turn 0 into 1, -1 into 2, -2 into 3 etc..
        coord = (node_id * -1) + 1
        node = JSONObject({
            'id' : node_id,
            'name' : node_name,
            'description' : "A node representing a water resource",
            'layout'      : None,
            'x' : 10 * coord,
            'y' : 10 * coord -1,
            'attributes' : attributes,
        })

        return node

    def build_network(self, project_id=None, num_nodes=10, new_proj=True, map_projection='EPSG:4326'):
        start = datetime.datetime.now()
        if project_id is None:
            proj_name = None

            if new_proj is True:
                proj_name = "Test Project @ %s"%(datetime.datetime.now())
                project_id = self.create_project(name=proj_name).id
            else:
                project_id = project_id

        LOG.debug("Project creation took: %s"%(datetime.datetime.now()-start))
        start = datetime.datetime.now()

        template = self.create_template()

        LOG.debug("Attribute creation took: %s"%(datetime.datetime.now()-start))
        start = datetime.datetime.now()

        # Put an attribute on a group
        group_ra = JSONObject(dict(
            ref_id = None,
            ref_key = 'GROUP',
            attr_is_var = 'N',
            attr_id = template.templatetypes[2].typeattrs[0].attr_id,
            id = -1
        ))
        group_attrs = [group_ra]

        nodes = []
        links = []

        prev_node = None
        ra_index = 2

        network_type = template.templatetypes[0]
        node_type = template.templatetypes[1]
        link_type = template.templatetypes[2]
        for n in range(num_nodes):
            node = self.create_node(n*-1, node_name="Node %s"%(n))

            #From our attributes, create a resource attr for our node
            #We don't assign data directly to these resource attributes. This
            #is done when creating the scenario -- a scenario is just a set of
            #data for a given list of resource attributes.
            node_ra1 = JSONObject(dict(
                ref_key = 'NODE',
                ref_id = None,
                attr_id = node_type.typeattrs[0].attr_id,
                id = ra_index * -1,
                attr_is_var = 'N',
            ))
            ra_index = ra_index + 1
            node_ra2 = JSONObject(dict(
                ref_key = 'NODE',
                ref_id = None,
                attr_id = node_type.typeattrs[1].attr_id,
                id = ra_index * -1,
                attr_is_var = 'Y',
            ))
            ra_index = ra_index + 1
            node_ra3 = JSONObject(dict(
                ref_key = 'NODE',
                ref_id = None,
                attr_id = node_type.typeattrs[2].attr_id,
                id = ra_index * -1,
                attr_is_var = 'N',
            ))
            ra_index = ra_index + 1
            node_ra4 = JSONObject(dict(
                ref_key = 'NODE',
                ref_id = None,
                attr_id = node_type.typeattrs[3].attr_id,
                id = ra_index * -1,
                attr_is_var = 'N',
            ))
            ra_index = ra_index + 1

            node.attributes = [node_ra1, node_ra2, node_ra3, node_ra4]

            type_summary = JSONObject(dict(
                template_id = template.id,
                template_name = template.name,
                id = node_type.id,
                name = node_type.name
            ))

            type_summary_arr = [type_summary]

            node.types = type_summary_arr

            nodes.append(node)

            if prev_node is not None:
                #Connect the two nodes with a link
                link = self.create_link(
                    n*-1,
                    node['name'],
                    prev_node['name'],
                    node['id'],
                    prev_node['id'])

                link_ra1 = JSONObject(dict(
                    ref_id = None,
                    ref_key = 'LINK',
                    id = ra_index * -1,
                    attr_id = link_type.typeattrs[0].attr_id,
                    attr_is_var = 'N',
                ))
                ra_index = ra_index + 1
                link_ra2 = JSONObject(dict(
                    ref_id = None,
                    ref_key = 'LINK',
                    attr_id = link_type.typeattrs[1].attr_id,
                    id = ra_index * -1,
                    attr_is_var = 'N',
                ))
                ra_index = ra_index + 1
                link_ra3 = JSONObject(dict(
                    ref_id = None,
                    ref_key = 'LINK',
                    attr_id = link_type.typeattrs[2].attr_id,
                    id = ra_index * -1,
                    attr_is_var = 'N',
                ))
                ra_index = ra_index + 1

                link.attributes = [link_ra1, link_ra2, link_ra3]
                if link['id'] % 2 == 0:
                    type_summary_arr = []
                    type_summary = JSONObject()
                    type_summary.template_id = template.id
                    type_summary.template_name = template.name
                    type_summary.id = link_type.id
                    type_summary.name = link_type.name

                    type_summary_arr.append(type_summary)

                    link.types = type_summary_arr
                links.append(link)

            prev_node = node

        #A network must contain an array of links. In this case, the array

        LOG.debug("Making nodes & links took: %s"%(datetime.datetime.now()-start))
        start = datetime.datetime.now()

        #Create the scenario
        scenario = JSONObject()
        scenario.id = -1
        scenario.name = 'Scenario 1'
        scenario.description = 'Scenario Description'
        scenario.layout = json.dumps({'app': ["Unit Test1", "Unit Test2"]})
        scenario.start_time = datetime.datetime.now()
        scenario.end_time = scenario.start_time + datetime.timedelta(hours=1)
        scenario.time_step = 1 # one second intervals.

        #Multiple data (Called ResourceScenario) means an array.
        scenario_data = []

        group_array = []
        group = JSONObject(dict(
            id = -1,
            name = "Test Group",
            description = "Test group description"
        ))

        group.attributes = group_attrs

        group_array.append(group)

        group_item_array = []
        group_item_1 = JSONObject(dict(
            ref_key = 'NODE',
            ref_id = nodes[0]['id'],
            group_id = group['id'],
        ))
        group_item_2 = JSONObject(dict(
            ref_key = 'NODE',
            ref_id = nodes[1]['id'],
            group_id = group['id'],
        ))
        group_item_array = [group_item_1, group_item_2]

        scenario.resourcegroupitems = group_item_array

        #This is an example of 3 diffent kinds of data

        #For Links use the following:
        #A simple string (Descriptor)
        #A multi-dimensional array.

        #For nodes, use the following:
        #A time series, where the value may be a 1-D array

        nodes[0].attributes
        for n in nodes:
            for na in n.attributes:
                if na.get('attr_is_var', 'N') == 'N':
                    if na['attr_id'] == node_type.typeattrs[0].attr_id:
                        #less than 10 and with 1 decimal place,
                        #as per the restriction in the template
                        dataset = self.create_scalar(na, 1.1, unit='cm^3')
                    elif na['attr_id'] == node_type.typeattrs[2].attr_id:
                        #incorrect unit to test the validation
                        dataset = self.create_timeseries(na, 'cm^3')
                    elif na['attr_id'] == node_type.typeattrs[3].attr_id:
                        dataset = self.create_dataframe(na, unit='m^3 s^-1')
                elif na.get('attr_is_var', 'Y') == 'Y':
                    if na['attr_id'] == node_type.typeattrs[1].attr_id:
                        # correct unit (speed)
                        dataset = self.create_scalar(na, unit='m s^-1')
                scenario_data.append(dataset)
        count = 0
        for l in links:
            for na in l.attributes:
                if na['attr_id'] == link_type.typeattrs[0].attr_id:
                    array = self.create_array(na)
                    scenario_data.append(array)
                    count = count + 1
                elif na['attr_id'] == link_type.typeattrs[1].attr_id:
                    descriptor = self.create_descriptor(na)
                    scenario_data.append(descriptor)
                    count = count + 1

        grp_timeseries = self.create_timeseries(group_attrs[0], 'cm^3')#incorrect unit to test validation

        scenario_data.append(grp_timeseries)

        #Set the scenario's data to the array we have just populated
        scenario.resourcescenarios = scenario_data

        #A network can have multiple scenarios, so they are contained in
        #a scenario array
        scenario_array = []
        scenario_array.append(scenario)

        LOG.debug("Scenario definition took: %s"%(datetime.datetime.now()-start))
        #This can also be defined as a simple dictionary, but I do it this
        #way so I can check the value is correct after the network is created.
        layout = dict(
            color = 'red',
            shapefile = 'blah.shp'
        )

        node_array = nodes
        link_array = links

        net_attr = self.create_attribute("net_attr_b", dimension='Pressure')

        ra_index = ra_index + 1
        net_ra_notmpl = JSONObject(dict(
            ref_id = None,
            ref_key = 'NETWORK',
            attr_is_var = 'N',
            attr_id = net_attr.id,
            id = ra_index*-1
        ))
        ra_index = ra_index + 1
        net_ra_tmpl = JSONObject(dict(
            ref_id = None,
            ref_key = 'NETWORK',
            attr_is_var = 'N',
            attr_id = network_type.typeattrs[0].attr_id,
            id = ra_index*-1
        ))
        net_attrs = [net_ra_notmpl, net_ra_tmpl]

        net_type_summary_arr = []
        net_type_summary = JSONObject(dict(
            template_id = template.id,
            template_name = template.name,
            id = network_type.id,
            name = network_type.name,
        ))
        net_type_summary_arr.append(net_type_summary)

        network = JSONObject(dict(
            name = 'Network @ %s'%datetime.datetime.now(),
            description = 'Test network with 2 nodes and 1 link',
            project_id = project_id,
            links = link_array,
            nodes = node_array,
            layout = layout,
            scenarios = scenario_array,
            resourcegroups = group_array,
            projection = map_projection,
            attributes = net_attrs,
            types = net_type_summary_arr,
        ))

        return network

    def create_network_with_data(self, project_id=None, num_nodes=10,
                                 ret_full_net=True, new_proj=False,
                                 map_projection='EPSG:4326'):
        """
            Test adding data to a network through a scenario.
            This test adds attributes to one node and then assigns data to them.
            It assigns a descriptor, array and timeseries to the
            attributes node.
        """

        network = self.build_network(project_id, num_nodes, new_proj=new_proj,
                                   map_projection=map_projection)

        #LOG.debug(network)
        start = datetime.datetime.now()
        LOG.info("Creating network...")
        response_network_summary = JSONObject(self.client.add_network(network))

        #Add an attribute scoped to this network, and an associated resource attribute
        net_scoped_attr = self.create_attribute("Network Scoped Attr", dimension=None, network_id=response_network_summary.id)
        self.client.add_resource_attribute('NETWORK', response_network_summary.id, net_scoped_attr.id, False)

        LOG.info("Network Creation took: %s"%(datetime.datetime.now()-start))
        if ret_full_net is True:
            LOG.info("Fetching new network...:")
            start = datetime.datetime.now()
            net = self.client.get_network(response_network_summary.id, include_data=True)
            response_net = JSONObject(net)
            LOG.info("Network Retrieval took: %s"%(datetime.datetime.now()-start))
            self.check_network(network, response_net)

            return response_net
        else:
           return response_network_summary

    def create_network_with_extra_group(self, project_id=None,
                                        num_nodes=10,
                                        ret_full_net=True,
                                        new_proj=False,
                                        map_projection='EPSG:4326'):
        """
            Test adding data to a network through a scenario.
            This test adds attributes to one node and then assigns data to them.
            It assigns a descriptor, array and timeseries to the
            attributes node.
        """
        network = self.create_network_with_data(
                                 project_id=project_id,
                                 num_nodes=num_nodes,
                                 ret_full_net=ret_full_net,
                                 new_proj=new_proj,
                                 map_projection=map_projection)

        group = JSONObject({})
        group.network_id = network.id
        group.id = -1
        group.name = 'test new group'
        group.description = 'test new group'

        template_id = network.types[0].template_id
        template = JSONObject(self.client.get_template(template_id))

        type_summary_arr = []

        type_summary = JSONObject({})
        type_summary.id = template.id
        type_summary.name = template.name
        type_summary.id = template.templatetypes[2].id
        type_summary.name = template.templatetypes[2].name

        type_summary_arr.append(type_summary)

        group.types = type_summary_arr

        new_group = self.client.add_group(network.id, group)

        group_attr_ids = []
        for resource_attr in new_group.attributes:
            group_attr_ids.append(resource_attr.attr_id)

        updated_network = self.client.get_network(network.id)

        return updated_network

    def create_network_with_child_scenario(self, project_id=None,
                                        num_nodes=10,
                                        ret_full_net=True,
                                        new_proj=False,
                                        map_projection='EPSG:4326',
                                        levels=2):
        """
            Create a network with two scenarios -- one with data in it, and its child containing
            none of its own data.
        """
        network = self.create_network_with_data(
                                 project_id=project_id,
                                 num_nodes=num_nodes,
                                 ret_full_net=ret_full_net,
                                 new_proj=new_proj,
                                 map_projection=map_projection)
        parent_scenario = network.scenarios[0]
        parent_scenario_id = parent_scenario.id
        scenario_count = len(network.scenarios) + 1
        for level in range(levels-1):
            new_scenario_j = self.client.create_child_scenario(parent_scenario_id, "Scenario {0}".format(scenario_count))
            parent_scenario_id = new_scenario_j.id
            scenario_count = scenario_count + 1


        updated_network = JSONObject(self.client.get_network(network.id, include_data=True))

        return updated_network



    def get_network(self, network_id=None):
        """
            Get a network with all data.
        """
        network = JSONObject(self.client.get_network(network_id))
        return network



    def check_network(self, request_net, response_net):
        assert response_net.layout == request_net['layout']

        assert response_net.scenarios[0].created_by is not None

        for n in response_net.nodes:
            assert n.x is not None
            assert n.y is not None
            assert len(n.attributes) > 0

        before_times = []

        s = request_net['scenarios'][0]
        print(s['resourcescenarios'])
        ordered_rs_request = sorted(s['resourcescenarios'], key=lambda x: str(x.dataset.value))
        for rs0 in ordered_rs_request:

            if rs0.dataset.type == 'timeseries':
                val = json.loads(rs0.dataset.value)
                before_ts_times = list(val.values())[0].keys()
                before_times = []
                for t in before_ts_times:
                    try:
                        before_times.append(get_datetime(t))
                    except Exception as err:
                        before_times.append(t)

        after_times = []
        s = response_net.scenarios[0]
        ordered_rs_response = sorted(s.resourcescenarios, key=lambda x: str(x.dataset.value))

        for rs0 in ordered_rs_response:
            if rs0.dataset.type == 'timeseries':
                val = json.loads(rs0.dataset.value)
                after_ts_times = list(val.values())[0].keys()
                after_times = []
                for t in after_ts_times:
                    try:
                        after_times.append(get_datetime(t))
                    except Exception as err:
                        after_times.append(t)

        for d in after_times:
            assert d in before_times, f"{d} is not in {before_times}"


    def create_scalar(self, resource_attr, val=1.234, unit='m^3'):
        #with a resource attribute.

        dataset = Dataset(dict(
            id=None,
            type = 'scalar',
            name = 'Flow speed',
            unit_id =  self.get_unit(unit).id,
            hidden = 'N',
            value = val,
        ))

        scenario_attr = JSONObject(dict(
            attr_id=resource_attr.attr_id,
            resource_attr_id=resource_attr.id,
            dataset=dataset,
        ))

        return scenario_attr

    def create_descriptor(self, resource_attr, val="test"):
        #A scenario attribute is a piece of data associated
        #with a resource attribute.

        dataset = Dataset(dict(
            id=None,
            type = 'descriptor',
            name = 'Flow speed',
            unit_id = self.get_unit('m s^-1').id, # This does not match the type on purpose, to test validation
            hidden = 'N',
            value = val,
        ))

        scenario_attr = JSONObject(dict(
            attr_id = resource_attr.attr_id,
            resource_attr_id = resource_attr.id,
            dataset = dataset,
        ))

        return scenario_attr


    def create_timeseries(self, resource_attr, unit='m^3'):
        #A scenario attribute is a piece of data associated
        #with a resource attribute.
        #[[[1, 2, "hello"], [5, 4, 6]], [[10, 20, 30], [40, 50, 60]]]

        fmt = hydra_base.config.get('DEFAULT', 'datetime_format', "%Y-%m-%dT%H:%M:%S.%f000Z")

        t1 = datetime.datetime.now(datetime.timezone.utc)
        t2 = t1+datetime.timedelta(hours=1)
        t3 = t1+datetime.timedelta(hours=2)

        val_1 = [[[1, 2, "hello"], [5, 4, 6]], [[10, 20, 30], [40, 50, 60]], [[9, 8, 7],[6, 5, 4]]]
        val_2 = [1.0, 2.0, 3.0]

        val_3 = [3.0, None, None]

        ts_val = {"test_column": {t1.strftime(fmt): val_1,
                      t2.strftime(fmt): val_2,
                      t3.strftime(fmt): val_3}}

        metadata = {'created_by': 'Test user'}

        dataset = Dataset(dict(
            id=None,
            type = 'timeseries',
            name = 'my time series',
            unit_id = self.get_unit(unit).id, # This does not match the type on purpose, to test validation
            hidden = 'N',
            value = json.dumps(ts_val),
            metadata = metadata
        ))

        scenario_attr = JSONObject(dict(
            attr_id = resource_attr.attr_id,
            resource_attr_id = resource_attr.id,
            dataset = dataset,
        ))

        return scenario_attr

    def create_dataframe(self, resource_attr, name='Test Data Frame', dataframe_value=None, unit='m^3 s^-1'):
        #A scenario attribute is a piece of data associated
        #with a resource attribute.

        if dataframe_value is None:
            val_1 = 1
            val_2 = 2
            val_3 = 3

            dataframe_value = {"test_column": {'key1': val_1,
                        'key2': val_2,
                        'key3': val_3}}

        metadata = {'created_by': 'Test user'}

        dataset = Dataset(dict(
            id=None,
            type = 'dataframe',
            name = name,
            unit_id = self.get_unit(unit).id,
            hidden = 'N',
            value = json.dumps(dataframe_value),
            metadata = metadata
        ))

        scenario_attr = JSONObject(dict(
            attr_id = resource_attr.attr_id,
            resource_attr_id = resource_attr.id,
            dataset = dataset,
        ))

        return scenario_attr

    def create_array(self, resource_attr):
        #A scenario attribute is a piece of data associated
        #with a resource attribute.
        #[[1, 2, 3], [4, 5, 6], [7, 8, 9]]

        arr = json.dumps([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])

        metadata_array = json.dumps({'created_by': 'Test user'})

        dataset = Dataset(dict(
            id=None,
            type = 'array',
            name = 'my array',
            unit_id = self.get_unit('bar').id,
            hidden = 'N',
            value = arr,
            metadata = metadata_array,
        ))

        scenario_attr = JSONObject(dict(
            attr_id = resource_attr.attr_id,
            resource_attr_id = resource_attr.id,
            dataset = dataset,
        ))

        return scenario_attr


    def create_attributes(self):
        dimension = "Volumetric flow rate"
        attrs = []
        attrs.append(
            JSONObject({
                'name'  : "Multi-added Attr 1",
                'dimension_id' : self.client.get_dimension_by_name(dimension).id,
                'description' : "Attribute 1 from a test of adding multiple attributes",
            })
        )
        attrs.append(
            JSONObject({
                'name' : "Multi-added Attr 2",
                'dimension_id' : self.client.get_dimension_by_name(dimension).id,
                'description' : "Attribute 2 from a test of adding multiple attributes",
            })
        )

        existing_attrs = []
        new_attrs = []

        for a in attrs:
            LOG.info("Getting attribute %s, %s", a.name, a.dimension_id)

            attr = self.client.get_attribute_by_name_and_dimension(a.name,
                                                                   a.dimension_id)
            if attr is not None and attr != {}:
                # The attribute already exists
                existing_attrs.append(attr)
            else:
                # The attribute is new
                new_attrs.append(a)

        added_attrs = self.client.add_attributes(new_attrs)

        # The existing attributes summed with the inserted ones are the total of "attrs" array
        total_attrs = existing_attrs + added_attrs

        assert len(total_attrs) == len(attrs)

        for my_attr in total_attrs:
            assert len(list(filter(lambda x: x.description == my_attr.description, attrs))) > 0

        return attrs

    def create_attribute(self, name="Test attribute", dimension=None, network_id=None, project_id=None):
        dimension_id = None
        if dimension is not None:
            dimension_id = self.client.get_dimension_by_name(dimension).id

        attr_i = self.client.get_attribute_by_name_and_dimension(name, dimension_id, network_id, project_id)
        if attr_i is None or attr_i == {}:
            attr = JSONObject({
                'name'  : name,
                'dimension_id' : dimension_id,
                'description' : "Attribute description",
                'network_id' : network_id,
                'project_id': project_id
            })
            attr = JSONObject(self.client.add_attribute(attr))
        else:
            attr = JSONObject(attr_i)

        return attr

    def create_attributegroup(self, project_id, name=None, exclusive='N'):

        if name is None:
            name = "Attribute Group %s" % (datetime.datetime.now(),)

        newgroup = JSONObject({
            'project_id'  : project_id,
            'name'        : name,
            'description' : "A description of an attribute group",
            'layout'      : {"color": "green"},
            'exclusive'   : exclusive,
            })

        newgroup = self.client.add_attribute_group(newgroup)
        return newgroup

    def get_by_name(self, name, entity_list):
        """
            given a list of JSONObjects with a name attribute, return the JSONObject with the
            specified name
        """
        entity_dict = {}
        for e in entity_list:
            entity_dict[e.name] = e

        return entity_dict[name]


    def create_dataset(self):
        """
            Creates a dataset and adds it to the DB
        """
        arr = json.dumps([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
        metadata_array = {'created_by': 'Test user'}
        u = self.get_unit('bar')
        # dataset = Dataset(dict(
        #     id=None,
        #     type = 'array',
        #     name = 'my array',
        #     unit_id = u.id,
        #     hidden = 'N',
        #     value = arr,
        #     metadata = metadata_array,
        # ))
        return  self.client.add_dataset(
            'array',
            arr,
            u.id,
            metadata = json.dumps(metadata_array),
            name = 'my array',
            flush = True
        )
